/**********************************************************************
 **                                                                  **
 **                            calculate.cc                          **
 **                                                                  **
 **                                                                  **
 **  Thus file contains the functions for Calculate class, which     **
 **  deals with various calculations in eBug.                        **
 **                                                                  **
 **  Written by:  Yoichiro Endo (Mobile Robot Lab, Georgia Tech)     **
 **                                                                  **
 **  Copyright 2000 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: calculate.cc,v 1.17 2000/10/25 04:44:32 endo Exp $ */

#include "ebug.h"

#define  SC            0.1 // SENSOR CONSTANT
//#define  OC            0.4 // ORIENTATION CONSTANT (HServer)
#define  OC            0.4 // ORIENTATION CONSTANT (Simulation)
//#define  PC            0.4 // PROGRESSION CONSTANT (HServer)
#define  PC           0.05 // PROGRESSION CONSTANT (Simulation)
#define  HVC           5.0 // HSERVER VELOCITY CONSTANT
#define  AOV           170 // ANGLE 0F VIEW
#define  MIN_DIST        1 // MINIMUM DISTANCE
#define  MIN_VTE_AA    0.1 // MINIMUM VTE ANGULAR ACCELERATION
#define  REACHABLE      15

//-----------------------------------------------------------------------------
// Calculate is a class which is a collection of functions that deal with
// various kinds of calculations.
//-----------------------------------------------------------------------------
Calculate::Calculate()
{
}

//-----------------------------------------------------------------------------
// This function calculates each photo-sensor value of the robot.
//-----------------------------------------------------------------------------
double Calculate::PhotoSensorValue(t_robot *robot, t_photo_sensor *photo_sensor, int light_type)
{
  // Initilize intensity reading at the photo-sensor for all stimulus of this kind.
  double Ip = 0;

  // Get the light list.
  t_node *pl = lightList;

  while (pl->next != NULL)
  {
    // Get the an individual light (stimulus) information.
    t_light *light = (t_light *)pl->data;
    
    if (light->type == light_type)
    {
      // Initilize intensity reading at the photo-sensor for this stimulus.
      double Ipi = 0;

      // Get the photo-sensor location.
      double sx = photo_sensor->sensor_x + robot->x;
      double sy = photo_sensor->sensor_y + robot->y;
      rotatePoint(robot->theta, &sx, &sy, robot->cmx+robot->x, robot->cmy+robot->y);

      // Get the normal of the sensor surface.
      double pdir = photo_sensor->angle + robot->theta;
      while (pdir >= 180) pdir -= 360;
      while (pdir < -180) pdir +=360;

      // Get the angle of the rays from the stimulus.
      double sdir = atan2(light->y - sy, light->x - sx);
      sdir = sdir*180.0/M_PI - 90;
      while (sdir >= 180) sdir -= 360;
      while (sdir < -180) sdir +=360;

      // Get the angle between the normal of the sensor surface and the angle of the
      // rays from the stimulus.
      double theta_ps = sdir - pdir;
      while (theta_ps >= 180) theta_ps -= 360;
      while (theta_ps < -180) theta_ps +=360;

      if (fabs(theta_ps) < (AOV/2.0))
      {
	// Get the distance between the sensor and the stimulus.
	double Dps = sqrt(pow(sx - light->x, 2)+pow(sy - light->y, 2));
	if (Dps <= MIN_DIST) Dps = MIN_DIST;

	// Get the coefficient value.
	double kp = photo_sensor->sensitivity*SC;

	// Get the intensity of the stimulus.
	double Is = light->lum;

	// Calculate the intensity reading at the photo-sensor for this stimulus.
	Ipi = kp*Is*cos(theta_ps*(M_PI/180.0))*(1.0/Dps);
      }

      // Record the sensor reading.
      photo_sensor->value[light_type] = Ipi;

      // Sum this to get the intensity reading at the photo-sensor for all stimulus of this kind.
      Ip += Ipi;
    }

    pl = pl->next;
  }
  return Ip;
}

//-----------------------------------------------------------------------------
// This function updates the vector value for the specified type (Orientation
// Vector or Progression Vector).
//-----------------------------------------------------------------------------
void Calculate::Vectors(int vector_type)
{
  // Get the robot list.
  t_node *pr = robotList;
  while (pr->next != NULL)
  {
    // Get an individual robot from the list.
    t_robot *robot = (t_robot *)pr->data;

    for (int light_type = 0; light_type < NUM_TYPES; light_type++)
    {
      // Retrieve the particular vector list.
      t_node *vecList = robot->vectorList[vector_type][light_type];

      while (vecList->next != NULL)
      {
	// Get an individual vector.
	t_vector *vector = (t_vector *)vecList->data;

	switch (vector_type) {
	case ORIENTATION:
	  vector->value = OrientationVector(robot, vector->side, light_type);
	  break;
	case PROGRESSION:
	  vector->value = ProgressionVector(robot, vector->side, light_type);
	  break;
	}

	vecList = vecList->next;
      }
    }
    pr = pr->next;
  }
}

//-----------------------------------------------------------------------------
// This function calculates the Orientation Vector value for the robot.
//-----------------------------------------------------------------------------
double Calculate::OrientationVector(struct t_robot *robot, int side, int light_type)
{
  // Initialize the Orentation Vector value.
  double V_O = 0;

  // Get the photo-sensor list.
  t_node *photoList = robot->photo_sensorList;

  while (photoList->next != NULL)
  {
    // Get an individual sensor information.
    t_photo_sensor *photo_sensor = (t_photo_sensor *)photoList->data;
    photo_sensor->value[light_type] = 0;

    // Get the sensor reading at the photo sensor;
    double Ip = PhotoSensorValue(robot, photo_sensor, light_type);

    // Get the Orientation Need value.
    double O = robot->orientation_need[light_type];

    if (photo_sensor->side != side)
    {
      // Get the height of the Orientation Distribution node.
      double op = Ip*O;

      // Orientation Vector value is the total sum of the Orientation Distribution
      // for the other side of the sowbug's body.
      V_O += op;
    }

    photoList = photoList->next;
  }
  return V_O;
}

//-----------------------------------------------------------------------------
// This function calculates the Progression Vector value of the robot.
//-----------------------------------------------------------------------------
double Calculate::ProgressionVector(struct t_robot *robot, int side, int light_type)
{
  // Initialize the Progression Vector value.
  double V_P = 0;

  // Get the photo-sensor list.
  t_node *photoList = robot->photo_sensorList;

  while (photoList->next != NULL)
  {
    // Get an individual sensor information.
    t_photo_sensor *photo_sensor = (t_photo_sensor *)photoList->data;
    photo_sensor->value[light_type] = 0;

    // Get the sensor reading at the photo sensor;
    double Ip = PhotoSensorValue(robot, photo_sensor, light_type);

    // Get the Hypothesis value.
    double H = robot->hypothesis[light_type];

    if (photo_sensor->side != side)
    {
      // Get the height of the Progression Distribution node.
      double pp = Ip*H;

      // Progression Vector value is the total sum of the Progression Distribution
      // for the other side of the sowbug's body.
      V_P += pp;
    }

    photoList = photoList->next;
  }
  return V_P;
}

//-----------------------------------------------------------------------------
// This function calculates the angular velocity of the robot from the
// Orientation and Progression Vectors.
//-----------------------------------------------------------------------------
void Calculate::AngularSpeed(struct t_robot *robot)
{
  // Initialize the angular velocities.
  double omega_total = 0;
  double omega[NUM_TYPES];
  robot->last_angvel = robot->angvel;
  robot->angvel = 0;

  // If there is no stimulus, no reaction will be generated.
  if (lightList->data == NULL) return;

  // Get the coefficients. (Let's say the effect of the Progrression Vector for
  // turning is 50% of the Orientation Vector.)
  double kappa_omega = OC;
  double kappa_omega_ = OC/2.0;

  for (int light_type = 0; light_type < NUM_TYPES; light_type++)
  {
    // Retreive the Orientation Vector value.
    double V_O[2] = {0.0,0.0};
    t_node *O_vecList = robot->vectorList[ORIENTATION][light_type];

    while (O_vecList->next != NULL)
    {
      t_vector *O_vector = (t_vector *)O_vecList->data;

      V_O[O_vector->side] +=  O_vector->value;

      O_vecList = O_vecList->next;
    }
  
    // Retreive the Progression Vector value.
    double V_P[2] = {0.0,0.0};
    t_node *P_vecList = robot->vectorList[PROGRESSION][light_type];

    while (P_vecList->next != NULL)
    {
      t_vector *P_vector = (t_vector *)P_vecList->data;

      V_P[P_vector->side] +=  P_vector->value;

      P_vecList = P_vecList->next;
    }

    // Calculate the angular velocity for this stimulus.
    omega[light_type] = kappa_omega*(V_O[LEFT] - V_O[RIGHT]) + kappa_omega_*(V_P[LEFT] - V_P[RIGHT]);
  }

  // The total angular velocity is a sum of the angular velocities for all the
  // stimulus types.
  for (int light_type = 0; light_type < NUM_TYPES; light_type++)
    omega_total += omega[light_type];

  robot->angvel = omega_total;
}

//-----------------------------------------------------------------------------
// This function calculates the forward speed of the robot from the Progression
// Vector.
//-----------------------------------------------------------------------------
void Calculate::ForwardSpeed(t_robot *robot)
{
  // Initialize the forward velocities.
  double v_total = 0;;
  double v[NUM_TYPES];
  robot->progression = 0;

  // If there is no stimulus, no reaction will be generated.
  if (lightList->data == NULL) return;

  // Get the coefficients.
  double k_v = PC;

  for (int light_type = 0; light_type < NUM_TYPES; light_type++)
  {
    // Retreive the Progression Vector value.
    double V_P[2] = {0.0,0.0};
    t_node *P_vecList = robot->vectorList[PROGRESSION][light_type];

    while (P_vecList->next != NULL)
    {
      t_vector *P_vector = (t_vector *)P_vecList->data;

      V_P[P_vector->side] += P_vector->value;

      P_vecList = P_vecList->next;
    }
    
    // Calculate the forward velocity of thus stimulus.
    v[light_type] = k_v*((V_P[RIGHT] + V_P[LEFT])/2);
  }

  // The total forward velocity is a sum of the forward velocities for all the
  // stimulus types.
  for (int light_type = 0; light_type < NUM_TYPES; light_type++)
    v_total += v[light_type];

  robot->progression = v_total;
}

//-----------------------------------------------------------------------------
// This fuction calculates the Hypothesis value of the robot.
//-----------------------------------------------------------------------------
void Calculate::Hypothesis(void)
{
  // If there is no robot or stimulus, the Hypothesis value cannot be calculated.
  if ((robotList->data == NULL)||(lightList->data == NULL)) return;

  LightFunction light_function;

  // Get the robot list.
  t_node *pr = robotList;
  while (pr->data != NULL)
  {
    // Get an individual robot information.
    t_robot *robot = (t_robot *)pr->data;

    // Get the closest stimulus information.
    double closest_dist = BIGNUMBER;
    t_light *closest_light = light_function.closestLight(robot, &closest_dist);

    // If the closest stimulus is close enough for the robot to reach, update the
    // Hypothesis value.
    if (closest_dist < REACHABLE)
    {
      // Update the Hypothesis value based on the stimulus type.
      switch (closest_light->status) {
      case POSITIVE:
	robot->hypothesis[closest_light->type] += HYPOTHESIS_INC;
	break;
      case NEUTRAL:
	//robot->hypothesis[closest_light->type] -= 1;
	break;
      case NEGATIVE:
	robot->hypothesis[closest_light->type] -= HYPOTHESIS_INC;
	break;
      }

      // Make sure the value is within the range.
      if (robot->hypothesis[closest_light->type] > HYPOTHESIS_MAX)
	robot->hypothesis[closest_light->type] = HYPOTHESIS_MAX;
      if (robot->hypothesis[closest_light->type] < HYPOTHESIS_MIN)
	robot->hypothesis[closest_light->type] = HYPOTHESIS_MIN;
      
      // If the robot can eat the stimulus and it is close enough to reach, erase it.
      if (eat_lights)
      {
	LightFunction light_function;
	light_function.remove(closest_light->id);
      }
      else
      {
	closest_light->status = NEUTRAL;
      }
    }
    pr = pr->next;
  }

}

//-----------------------------------------------------------------------------
// This function calculates the Orientation Need value of the robot.
//-----------------------------------------------------------------------------
void Calculate::OrientationNeed(void)
{
  // If there is no robot or stimulus, the Orientation Need value cannot be calculated.
  if ((robotList->data == NULL)||(lightList->data == NULL)) return;

  LightFunction light_function;

  // Get the robot list.
  t_node *pr = robotList;
  while (pr->data != NULL)
  {
    // Get an individual robot information.
    t_robot *robot = (t_robot *)pr->data;

    for (int light_type = 0; light_type < NUM_TYPES; light_type++)
    {
      // If the robot is directly looking at the stimulus, the Orientation value
      // will be reduced to zero. Otherwise, it is the default value.
      if (light_function.lookingLightDirectly(robot, light_type))
	robot->orientation_need[light_type]  = 0;
      else
	robot->orientation_need[light_type] = ORIENTATION_NEED_DEFAULT_VALUE;
    }

    pr = pr->next;
  }

}
//-----------------------------------------------------------------------------
// This function calculates the Vicarious Trial and Error (VTE) of the robot.
//-----------------------------------------------------------------------------
void Calculate::VTE(void)
{
  // If the robot is not moving, there will be no VTE.
  if (!go) return;

  // Get the robot list.
  t_node *pr = robotList;
  while (pr->data != NULL)
  {
    // Get an individual robot information.
    t_robot *robot = (t_robot *)pr->data;

    // If the current and last angular velocities have same sign (+/-)
    // and the difference are large enough, VTE is counted.
    if ((signPN(robot->angvel) != signPN(robot->last_angvel)) &&
	(fabs(robot->angvel - robot->last_angvel) > MIN_VTE_AA))
    {
      ++robot->vte;
    }

    if (debug)
    {
      fprintf(stderr, "VTE = %d %d %d %f\n", robot->vte, signPN(robot->angvel),
	      signPN(robot->last_angvel), fabs(robot->angvel - robot->last_angvel));
    }

    pr = pr->next;
  }
}

//-----------------------------------------------------------------------------
// This function calculates the input angle in 0 to 360 range.
//-----------------------------------------------------------------------------
void Calculate::convertTheta(double *theta)
{
  while (*theta <= 0) *theta += 360;
  *theta = fmod(*theta,360);
}

//-----------------------------------------------------------------------------
// eBug and HServer deal with objects in different coordinate systems. This
// function converts one to the other.
//-----------------------------------------------------------------------------
void Calculate::convertCoordinate(double *x, double *y, double *theta, int in, int out)
{
  double factor, constant;

  // Converting from eBug to HServer 
  if ((in == EBUG) && (out == HSERVER))
  {
    factor = SIM_TO_REAL;
    constant = - 90;
  }
  // Converting from HServer to eBug
  else if ((in == HSERVER) && (out == EBUG))
  {
    factor =  1/SIM_TO_REAL;
    constant = + 90;
  }

  *x *= factor;
  *y *= factor;
  *theta += constant;

}

//-----------------------------------------------------------------------------
// The units of eBug is in pixel while HServer calculates everything in meter.
// This function convers to the other.
//-----------------------------------------------------------------------------
void Calculate::convertVelocity(double *vel, int type, int in, int out)
{
  if (delta_time <= SMALLNUMBER)
  {
    fprintf(stderr, "Error: deleta_time too small\n");
    return;
  }

  double factor;

  // Converting from eBug to HServer 
  if ((in == EBUG) && (out == HSERVER))
  {
    factor = HVC;
    if (type == LINEAR) factor *= SIM_TO_REAL;
  }
  // Converting from HServer to eBug
  else if ((in == HSERVER) && (out == EBUG))
  {
    factor = 1/HVC;
    if (type == LINEAR) factor /= SIM_TO_REAL;
  }

  *vel *= factor;
}

//-----------------------------------------------------------------------------
// This function labels the input value to be either positive or negative.
//-----------------------------------------------------------------------------
int Calculate::signPN(double value)
{
  if (fabs(value) <= SMALLNUMBER) return POSITIVE; 
  if ((value/fabs(value) >= 0)) return POSITIVE;
  else return NEGATIVE;
}
//-----------------------------------------------------------------------------
// This function calculates the difference between two angles.
//-----------------------------------------------------------------------------
double Calculate::diffAngle(double ang1, double ang2)
{
  double diff;
  if (ang1>ang2) diff = ang1-ang2;
  else diff = ang2-ang1;
  if (diff>180) diff = 360 - diff;
  return(diff);
}

//-----------------------------------------------------------------------------
// This function checks whether the input angle is within the field of view.
//-----------------------------------------------------------------------------
int Calculate::inFieldOfView(double smin, double smax, double ang)
{
  if (smin<smax) {
    if ((ang<=smax)&&(ang>=smin)) return(TRUE);
  }
  else {
    if (smax>ang) return(TRUE);
    if (smin<ang) return(TRUE);
  }
  return(FALSE);
}
//-----------------------------------------------------------------------------
// This function rotates an input point (x, y) around the center point (cx, cy).
//-----------------------------------------------------------------------------
void Calculate::rotatePoint(double theta, double *x, double *y, double cx, double cy)
{
  double stheta,ctheta,tmpx;
  theta = theta * M_PI / 180;
  stheta = sin(theta);
  ctheta = cos(theta);

  *x -= cx;
  *y -= cy;
  tmpx = ctheta*(*x)-stheta*(*y);
  *y = stheta*(*x)+ctheta*(*y);
  *x = tmpx;
  *x += cx;
  *y += cy;
}

//-----------------------------------------------------------------------------
// This function was a part of the original code to simulate the Braitenberg
// Vehicles.
//-----------------------------------------------------------------------------
void Calculate::BVsensorValues(void)
{
  t_node *pr = robotList;
  while (pr->next != NULL)
  {
    t_robot *robot = (t_robot *)pr->data;

    t_node *ps = robot->sensorList;
    if (ps->data == NULL) return;

    while (ps->next != NULL)
    {
      t_sensor *sensor = (t_sensor *)ps->data;

      double sx = sensor->x + robot->x;
      double sy = sensor->y + robot->y;
      rotatePoint(robot->theta,&sx,&sy,robot->cmx+robot->x,robot->cmy+robot->y);

      double shead = sensor->angle + robot->theta;
      convertTheta(&shead);
      double smin = (shead - sensor->fov/2.0);
      convertTheta(&smin);

      double smax = (shead + sensor->fov/2.0);
      convertTheta(&smax);

      sensor->stimulus = 0;

      t_node *pl = lightList;
      if (pl->data == NULL) return;
	  
      while (pl->next != NULL)
      {
	t_light *light = (struct t_light *)pl->data;
	double lhead = atan2(light->y-sy,light->x-sx);
	lhead = lhead * 180 / M_PI - 90;
	convertTheta(&lhead);

	double ldist = sqrt(pow(sx-light->x,2)+pow(sy-light->y,2));

	if (inFieldOfView(smin,smax,lhead)) 
	  sensor->stimulus += sensor->sensitivity*light->lum*(1/ldist);

	pl = pl->next;
      }
      ps = ps->next;
    }
    pr = pr->next;
  }
}

//-----------------------------------------------------------------------------
// This function was a part of the original code to simulate the Braitenberg
// Vehicles.
//-----------------------------------------------------------------------------
void Calculate::BVmotorForce(struct t_robot *robot)
{

  struct t_sensor *sensor;
  struct t_motor *motor;
  struct t_neuron *neuron;
  struct t_node *pn,*pm,*ps;

  BVclearMotorForce(robot);

  pn = robot->neuronList;
  if (pn->data == NULL) return;
  do {

    neuron = (struct t_neuron *)pn->data;
    pm = robot->motorList;
    if (pm->data == NULL) return;
    do {
      motor = (struct t_motor *)pm->data;
      pm = pm->next;
    } while (motor->id != neuron->motorId);

    ps = robot->sensorList;
    if (ps->data == NULL) return;
    do {
      sensor = (struct t_sensor *)ps->data;
      ps = ps->next;
    } while (sensor->id != neuron->sensorId);

    motor->force += sensor->stimulus*motor->strength;
    if (neuron->neuronType == NT_THRESHHOLD) 
      {
	if (motor->force > motor->maxForce) motor->force = motor->maxForce - motor->force;
	if (motor->force < 0) motor->force = 0;
      }
    if (neuron->neuronType == NT_THRESHHOLDNEG)
      {
	if (motor->force > motor->maxForce) motor->force = motor->maxForce - motor->force;
	if (motor->force < -motor->maxForce) motor->force = -motor->maxForce;
      }
    if (motor->force > motor->maxForce) motor->force = motor->maxForce;
    if (neuron->neuronType == NT_NEGATIVE) motor->force = -motor->force;
    pn = pn->next;
  } while (pn->next != NULL);
}  

//-----------------------------------------------------------------------------
// This function was a part of the original code to simulate the Braitenberg
// Vehicles.
//-----------------------------------------------------------------------------
void Calculate::BVclearMotorForce(struct t_robot *robot)
{
  struct t_node *p;
  struct t_motor *motor;
  
  p = robot->motorList;
  if (p->data == NULL) return;
  do {
    motor = (struct t_motor *)p->data;
    motor->force = 0;
    p = p->next;
  } while (p->next != NULL);
}

//-----------------------------------------------------------------------------
// This function was a part of the original code to simulate the Braitenberg
// Vehicles.
//-----------------------------------------------------------------------------
void Calculate::BVtotalForce(struct t_robot *robot, double *force, double *angle)
{
  struct t_motor *motor;
  struct t_node *p;
  double mhead, mxf = 0, myf = 0;

  p = robot->motorList;
  if (p->data == NULL) return;
  do {
    motor = (struct t_motor *)p->data;
    mhead = motor->angle + robot->theta;
    mxf -= motor->force*sin(mhead*M_PI/180);
    myf += motor->force*cos(mhead*M_PI/180);
    p = p->next;
  } while (p->next != NULL);
  *force = sqrt(pow(mxf,2)+pow(myf,2));
  if ((myf!=0)&&(mxf!=0)) *angle = atan2(myf,mxf) * 180 / M_PI - 90;
  convertTheta(angle);
}

//-----------------------------------------------------------------------------
// This function was a part of the original code to simulate the Braitenberg
// Vehicles.
//-----------------------------------------------------------------------------
void Calculate::BVtorque(struct t_robot *robot, double *angvel)
{
  struct t_motor *motor;
  struct t_node *p;
  double angtocm,angtocmr,torqueAng,f2tAng,torque,disttocm;

  *angvel = 0;
  p = robot->motorList;
  if (p->data == NULL) return;
  do {
    motor = (struct t_motor *)p->data;
    angtocm = atan2(robot->cmy-motor->y,robot->cmx-motor->x) * 180 / M_PI - 90;
    convertTheta(&angtocm);
    disttocm = sqrt(pow(robot->cmy-motor->y,2)+pow(robot->cmx-motor->x,2));
    angtocmr = angtocm - 180;
    convertTheta(&angtocmr);
    if (inFieldOfView(angtocmr,angtocm,motor->angle)) torqueAng = angtocm - 90;
    else torqueAng = angtocm + 90;
    convertTheta(&torqueAng);
    f2tAng = diffAngle(motor->angle,torqueAng);
    torque = disttocm*cos(f2tAng*M_PI/180)*motor->force;
//    if (torque > robot->maxTorque) torque = robot->maxTorque;
    if (!inFieldOfView(angtocmr,angtocm,torqueAng)) torque = -torque;
    *angvel += torque;
    /*printf("tq: %f %f %f %f %f %f\n",angtocm,angtocmr,torqueAng,f2tAng,torque,*angvel);*/
    p = p->next;
  } while (p->next != NULL);
}
  
/**********************************************************************
 * $Log: calculate.cc,v $
 * Revision 1.17  2000/10/25 04:44:32  endo
 * Code cleaned up for the distribution.
 *
 * Revision 1.16  2000/08/19 20:43:39  endo
 * Slight changes due to the hardware interface.
 *
 * Revision 1.15  2000/08/10 06:16:05  endo
 * convertCoordinate, convertVelocity etc.
 *
 * Revision 1.14  2000/08/07 21:00:02  endo
 * Reset feature added.
 *
 * Revision 1.13  2000/08/06 07:28:14  endo
 * Two light types implemented successfully.
 *
 * Revision 1.12  2000/07/31 05:31:04  endo
 * Hypothesis feature improved. Orientation Need feature added.
 *
 * Revision 1.11  2000/07/30 23:50:23  endo
 * cleaned up.
 *
 * Revision 1.10  2000/07/29 22:18:15  endo
 * Reward stuff implemented.
 *
 * Revision 1.9  2000/04/30 23:56:03  endo
 * Finally, Internal Graph seems to be working correctly.
 *
 * Revision 1.8  2000/04/30 12:41:34  endo
 * Internal Graph implemented. Not perfect, though.
 *
 * Revision 1.7  2000/04/30 08:03:03  endo
 * photo_sensor implemented.
 *
 * Revision 1.6  2000/04/30 06:39:11  endo
 * Vector values are now function of the angle between
 * lights and sensor, too.
 *
 * Revision 1.5  2000/04/29 09:17:44  endo
 * Progression Vector working.
 *
 * Revision 1.4  2000/04/28 21:33:30  endo
 * Orientation vector seems to be right.
 *
 * Revision 1.3  2000/04/28 18:01:54  endo
 * vector.value works now.
 *
 * Revision 1.2  2000/04/28 09:18:33  endo
 * Vectors, OrientationVector, ProgressionVector added.
 * vector.value only show up in right side...
 *
 * Revision 1.1  2000/03/27 00:50:51  endo
 * Initial revision
 *
 *
 **********************************************************************/
