/**********************************************************************
 **                                                                  **
 **                             light.cc                             **
 **                                                                  **
 **                                                                  **
 **  This file contains the functions for LightFunction class,       **
 **  which deals the stimuli (lights).                               **
 **                                                                  **
 **  Written by:  Yoichiro Endo (Mobile Robot Lab, Georgia Tech)     **
 **                                                                  **
 **  Copyright 2000 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: light.cc,v 1.8 2000/10/25 04:45:27 endo Exp $ */

#include "ebug.h"

#define DEFAULT_INC 2
#define MIN_RADIUS  10
#define MAX_RADIUS  40
#define DAOV        5 // DIRECT ANGLE OF VIEW

//-----------------------------------------------------------------------------
// LightFunction is a class which is a collection of functions that deal with
// stimuli (lights) relatelated issue.
//-----------------------------------------------------------------------------
LightFunction::LightFunction()
{
}

//-----------------------------------------------------------------------------
// This function initializes the stimuli in the display.
//-----------------------------------------------------------------------------
void LightFunction::initialize(void)
{
  ListUtility list_utility;
  
  list_utility.clearList(&lightList);

  lightId = 0;

  // Load default light
  if (load_default_light)
  {
    load(default_light);

    if (debug)
      fprintf(stderr, "The default lights \"%s\" load.\n", default_light);
  }

}

//-----------------------------------------------------------------------------
// This function checks the point is one of the stimulus regions.
//-----------------------------------------------------------------------------
int LightFunction::inLight(point pt, int *id)
{
  if (lightList->data == NULL) return(0);

  t_node *p = lightList;

  while (p->next != NULL)
  {
    t_light *light = (t_light *)p->data;
    int     lx = pt.x - light->x;
    int     ly = pt.y - light->y;
    double  dist = sqrt(lx*lx + ly*ly);

    if (dist <= light->radius)
    {
      *id = light->id;
      return(1);
    }
    p = p->next;
  }

  return(0);
}

//-----------------------------------------------------------------------------
// This function fines the closest stimulus to the robot.
//-----------------------------------------------------------------------------
t_light *LightFunction::closestLight(t_robot *robot, double *closest_dist)
{
  Calculate calculate;

  t_light *closest_light;
  t_node *ps = robot->photo_sensorList;

  if (ps->data == NULL) return NULL;
  while (ps->next != NULL)
  {
    t_photo_sensor *photo_sensor = (t_photo_sensor *)ps->data;
	  
    // Get the current photo sensor location.
    double sx = photo_sensor->sensor_x + robot->x;
    double sy = photo_sensor->sensor_y + robot->y;
    calculate.rotatePoint(robot->theta, &sx, &sy, robot->cmx+robot->x, robot->cmy+robot->y);

    t_node *pl = lightList;
    if (pl->data == NULL) return NULL;

    while  (pl->next != NULL)
    {
      t_light *light = (t_light *)pl->data;
      double dist = sqrt(pow((sx-light->x),2) + pow((sy-light->y),2));

      if (*closest_dist >= dist)
      {
	*closest_dist = dist;
	closest_light = light;
      }
      pl = pl->next;
    }
    ps = ps->next;
  }
  return closest_light;
}

//-----------------------------------------------------------------------------
// This function checks whether the robot is directly looking at the stimulus.
//-----------------------------------------------------------------------------
int LightFunction::lookingLightDirectly(t_robot *robot, int light_type)
{
  if (lightList->data == NULL) return FALSE;

  t_node *pl = lightList;
  while (pl->next != NULL)
  {
    t_light *light = (t_light *)pl->data;

    if (light->type == light_type)
    {
      double lx = (double)light->x;
      double ly = (double)light->y;
      double rx = robot->x;
      double ry = robot->y;
      double angle = atan2(ly-ry, lx-rx)*180.0/M_PI - robot->theta + 90;
      while (angle >= 180) angle -= 360;
      while (angle < -180) angle += 360;

      if (fabs(angle) < DAOV) return TRUE;
    }
    pl = pl->next;
  }
  return FALSE;
 }

//-----------------------------------------------------------------------------
// This function chages the status (positive, neutral, or negative) of the
// stimulus.
//-----------------------------------------------------------------------------
void LightFunction::changeStatus(int id, int status)
{
  if (lightList->data == NULL) return;

  ScreenUtility screen_utility;
  if (lightList->data != NULL) screen_utility.drawLights(ERASE);

  ListUtility list_utility;
  t_node *new_lightList;
  list_utility.create(&new_lightList);

  while (lightList->next != NULL)
  {
    t_light *light = (t_light *)lightList->data;

    if (light->id == id)
    {
      if (status == TOGGLE)
      {
	light->status += 1;
	if (light->status > POSITIVE) light->status = NEGATIVE;
      }
      else
	light->status = status;
    }
    list_utility.push(&new_lightList,(int*)light);

    lightList = lightList->next;
  }
  list_utility.clearList(&lightList);
  lightList = new_lightList;

  if (lightList->data != NULL) screen_utility.drawLights(DRAW);
}

//-----------------------------------------------------------------------------
// This function checks the type (A or B) of the stimulus.
//-----------------------------------------------------------------------------
int LightFunction::lightType(int id)
{
  if (lightList->data == NULL) return -1;

  int light_type;

  t_node *p = lightList;
  while (p->next != NULL)
  {
    t_light *light = (t_light *)lightList->data;

    if (light->id == id)
    {
      light_type =  light->type;
      break;
    }

    p = p->next;
  }

  return light_type;
}

//-----------------------------------------------------------------------------
// This function changes the type (A or B) of the stimulus.
//-----------------------------------------------------------------------------
void LightFunction::changeType(int id)
{
  if (lightList->data == NULL) return;

  ScreenUtility screen_utility;
  if (lightList->data != NULL) screen_utility.drawLights(ERASE);

  ListUtility list_utility;
  t_node *new_lightList;
  list_utility.create(&new_lightList);

  while (lightList->next != NULL)
  {
    t_light *light = (t_light *)lightList->data;

    if (light->id == id)
    {
      light->type += 1;
      if (light->type >= NUM_TYPES) light->type = TYPE_A;
    }
    list_utility.push(&new_lightList,(int*)light);

    lightList = lightList->next;
  }
  list_utility.clearList(&lightList);
  lightList = new_lightList;

  if (lightList->data != NULL) screen_utility.drawLights(DRAW);
}

//-----------------------------------------------------------------------------
// This function changes the size and the intensity of the stimulus.
//-----------------------------------------------------------------------------
void LightFunction::changeSize(int id)
{
  if (lightList->data == NULL) return;

  ScreenUtility screen_utility;
  if (lightList->data != NULL) screen_utility.drawLights(ERASE);

  t_node *p = lightList;

  while (p->next != NULL)
  {
    t_light *light = (t_light *)p->data;

    if (light->id == id)
    {
      light->radius += DEFAULT_INC;
      light->lum = light->radius;
      if (light->radius>MAX_RADIUS)
      {
	light->radius = MIN_RADIUS;
	light->lum = light->radius;
      }
    }
    p = p->next;
  }

  if (lightList->data != NULL) screen_utility.drawLights(DRAW);
}

//-----------------------------------------------------------------------------
// This function updates the age of the stimuli.
//-----------------------------------------------------------------------------
void LightFunction::updateLightAge(void)
{
  if (lightList->data == NULL) return;

  t_node *p = lightList;
  while (p->next != NULL)
  {
    t_light *light = (t_light *)p->data;

    light->age += delta_time;

    p = p->next;
  }
}

//-----------------------------------------------------------------------------
// This function deletes the stimuli that are older than they are supposed to
// be.
//-----------------------------------------------------------------------------
void LightFunction::deleteOldLights(double max_old)
{
  if (lightList->data == NULL) return;

  t_node *p = lightList;
  while (p->next != NULL)
  {
    t_light *light = (t_light *)p->data;

    if (light->age > max_old)
      remove(light->id);

    p = p->next;
  }
}

//-----------------------------------------------------------------------------
// This function removes the stimulus from the list.
//-----------------------------------------------------------------------------
void LightFunction::remove(int id)
{
  if (lightList->data == NULL) return;

  ScreenUtility screen_utility;
  if (lightList->data != NULL) screen_utility.drawLights(ERASE);

  ListUtility list_utility;
  t_node *new_lightList;
  list_utility.create(&new_lightList);

  while (lightList->next != NULL)
  {
    t_light *light = (t_light *)lightList->data;

    if (light->id != id) 
      list_utility.push(&new_lightList,(int*)light);

    lightList = lightList->next;
  }
  list_utility.clearList(&lightList);
  lightList = new_lightList;

  if (lightList->data != NULL) screen_utility.drawLights(DRAW);
}

//-----------------------------------------------------------------------------
// This function adds a stimulus to the list and display in the point specified. 
//-----------------------------------------------------------------------------
void LightFunction::add(point pt)
{
  ListUtility        list_utility;

  t_light *tmpLight = (t_light *)malloc(sizeof(t_light));
  tmpLight->id = lightId++;
  tmpLight->x = pt.x;
  tmpLight->y = pt.y;
  tmpLight->radius = MIN_RADIUS;
  tmpLight->lum = tmpLight->radius;
  tmpLight->status = NEUTRAL;
  tmpLight->type = TYPE_A;
  tmpLight->age = 0.0;
  list_utility.push(&lightList,(int*)tmpLight);
}

//-----------------------------------------------------------------------------
// This function loads the stimuli to the display from the specified file.
//-----------------------------------------------------------------------------
void LightFunction::load(char *filename)
{
  ListUtility        list_utility;

  FILE *fp = fopen(filename,"r");
  if (fp==NULL)
  {
    fprintf(stderr, "Error: %s cannot be opened.\n", filename);
    return;
  }

  while (1)
  {
    t_light *light = (t_light *)malloc(sizeof(t_light));
    fscanf(fp, "%d %d %d %d %d %d %d", &(light->id), &(light->x), &(light->y),
	   &(light->lum), &(light->radius), &(light->status), &(light->type));
	  
    if (light->id != -1)
    {
      list_utility.push(&lightList,(int*)light);
      lightId++;
    }
    else break;
  }

  fclose(fp);
}
  
//-----------------------------------------------------------------------------
// This function saves the stimuli in the display into a specified file.
//-----------------------------------------------------------------------------
void LightFunction::save(void)
{
  FILE *fp;
  struct t_node *p;
  struct t_light *light;
  char filename[100];

  printf("File name: ");
  scanf("%s",filename);
  fp = fopen(filename,"w");

  if (fp == NULL) return;

  p = lightList;
  if (p->data == NULL) return;

  do {
    light = (struct t_light *)p->data;
    fprintf(fp, "%d %d %d %d %d %d %d\n", light->id, light->x, light->y, light->lum,
	    light->radius, light->status, light->type);
    p = p->next;
  } while (p->next != NULL);

  fprintf(fp,"-1 0 0 0 0 0 0\n");
  fclose(fp);

  fprintf(stderr, "Lights saved in \"%s\".\n", filename);
}
  
/**********************************************************************
 * $Log: light.cc,v $
 * Revision 1.8  2000/10/25 04:45:27  endo
 * Code cleaned up for the distribution.
 *
 * Revision 1.7  2000/08/19 20:46:29  endo
 * Slight changes due to the hardware interface.
 *
 * Revision 1.6  2000/08/07 21:00:39  endo
 * Reset feature added.
 *
 * Revision 1.5  2000/08/06 07:27:55  endo
 * Two light types implemented successfully.
 *
 * Revision 1.4  2000/07/31 05:31:33  endo
 * Hypothesis feature improved. Orientation Need feature added.
 *
 * Revision 1.3  2000/07/29 22:18:29  endo
 * Some clean up.
 *
 * Revision 1.2  2000/03/26 08:35:46  endo
 * Finally, the simulation works as it supposed to be.
 *
 * Revision 1.1  2000/03/26 07:58:13  endo
 * Initial revision
 *
 *
 **********************************************************************/
