/**********************************************************************
 **                                                                  **
 **                              main.cc                             **
 **                                                                  **
 **                                                                  **
 **  This file contains the main function and main-loop function of  **
 **  eBug.                                                           **
 **                                                                  **
 **  Written by:  Yoichiro Endo (Mobile Robot Lab, Georgia Tech)     **
 **                                                                  **
 **  Copyright 2000 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: main.cc,v 1.5 2000/10/28 00:15:30 endo Exp endo $ */


#include "ebug.h"

//-----------------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------------
Display            *display;
Widget             top_level;
Widget             main_window, scrolled_window;
Widget             work_area, drawing_area;
Widget             command_bar, menu_bar;
Widget             load_lights_button, run_pause_button, reset_button, quit_button;
Widget             hypothesis_window_shell, hypothesis_window, hypothesis_scale[NUM_TYPES];
Widget             orientation_need_window_shell, orientation_need_window, orientation_need_scale[NUM_TYPES];
Pixmap             drawing_area_pixmap, plane;
Colormap           ebug_colormap;
GC                 gc;
GC                 blueline1gc, blueline3gc;
GC                 lightblueline1gc, lightblueline3gc;
GC                 redline1gc, redline3gc;
GC                 pinkline1gc, pinkline3gc;
GC                 trailgc;
GC                 sensorgc;
GC                 lightgc[NUM_TYPES];
GC                 vectorgc[NUM_VECTORS][NUM_TYPES];
GC                 graphgc[NUM_VECTORS][NUM_TYPES];
GC                 theGC;
GC                 xorGC;
GC                 red_xorGC;
Pixel              red_pixel, pink_pixel, blue_pixel, lightblue_pixel, green_pixel;
Pixel              lightcolor_pixel[NUM_TYPES];
Pixel              vector_pixel[NUM_VECTORS][NUM_TYPES];
Pixel              graph_pixel[NUM_VECTORS][NUM_TYPES];
int                screen_num;
XtWorkProc         workproc[MAX_WORKPROCS];
XtAppContext       app;
XEvent             event;

struct t_node      *robotnameList, *robotList, *lightList;
unsigned int       width, height;

double             birth_time;

//-----------------------------------------------------------------------------
// Default setups
//-----------------------------------------------------------------------------
int run_type                 = SIMULATION;
unsigned long event_mask     = ExposureMask|KeyPressMask|ButtonPressMask|StructureNotifyMask;
bool show_trails             = TRUE;
bool load_default_robot      = TRUE;
char *default_robot          = "sowbug00.bot";
bool load_default_light      = TRUE;
char *default_light          = "light01.stm";
bool eat_lights              = TRUE;
bool debug                   = FALSE;
bool log_data                = FALSE;
char *log_filename           = "ebug.log";
char *program_name           = "eBug";
char *binary_name            = "ebug";
char *version                = "1.0";
char *window_sub_title       = "Tolman's Schematic Sowbug";

//-----------------------------------------------------------------------------
// Initial Values
//-----------------------------------------------------------------------------
int  go                      = FALSE;
int  trial                   = 0;
int  lightId                 = 0;

//-----------------------------------------------------------------------------
// This is a main loop function of eBug. It will be called every time cycle
// to update the robot position.
//-----------------------------------------------------------------------------
void ebugMainLoop(void)
{
  ScreenUtility   screen_utility;
  TimeUtility     time_utility;
  LogUtility      log_utility;
  Calculate       calculate;
  RobotFunction   robot_function;
  HardwareCommand hardware_command;

  // Get the current time.
  double time_cycle_init = time_utility.currentTime();

  // Make sure there is a data on either/both the robot and light lists.
  if ((robotList->data != NULL)||(lightList->data != NULL))
  {
    // Run the simulation loop.
    while(1)
    {
      // Get the current time.
      double time_cycle_now = time_utility.currentTime();
      double delay = time_cycle_now - time_cycle_init;

      if (delay > UPDATE_CYCLE)
      {
	if (robotList->data != NULL)
	{
	  // Erase the robot first.
	  screen_utility.drawRobots(ERASE, PRIMARY);

	  // Erase lights		   
	  screen_utility.drawLights(ERASE);

          // Get the robot age.
	  time_utility.computeAge();

	  if (go)
	  {
	    // Update Hypothesis and Orientation Need values
	    calculate.Hypothesis();
	    calculate.OrientationNeed();
	  }

	  // Calculate Orientation and Progression vectors
	  calculate.Vectors(ORIENTATION);
	  calculate.Vectors(PROGRESSION);

	  if (go)
	  {
	    // Move the robot based on the vectors.
	    robot_function.driveRobot();
	    if (run_type == HSERVER) hardware_command.driveRobot();

	    // Calculate the Vicarious Trial and Error (VTE) value.
	    calculate.VTE();

	    // Record the data.
	    if (log_data) log_utility.recordPerformance();
	  }
	  else
	  {
	    if (run_type == HSERVER) hardware_command.stopRobot();
	  }

	  // If it is for the real robot, get the current information of the robot
	  // and lights.
	  if (run_type == HSERVER)
	  {
	    hardware_command.getCurrentRobotInfo();
	    hardware_command.getCurrentLightInfo();
	  }

	  // Draw the robot on the screen.
	  screen_utility.drawRobots(DRAW, PRIMARY);
	}
	       
	// Display misc information on the screen.
	screen_utility.displayAge();
	screen_utility.displayTrials();
	screen_utility.displayVTE();

	// Draw light(stimulus).
	screen_utility.drawLights(DRAW);

	// Update the Run/Pause button.
	if ((lightList->data == NULL) || (go && !robot_function.isMoving()))
	{
	  go = FALSE;
	  screen_utility.updateRunPauseButton();
	}

	// Update other windows.
	screen_utility.updateHypothesisWindow();
	screen_utility.updateOrientationNeedWindow();

	time_cycle_init = time_utility.currentTime();
      }

      if (XCheckMaskEvent(XtDisplay(main_window), event_mask, &event))
      {
	XtDispatchEvent (&event);
	return;
      }
    }
  }
}

//-----------------------------------------------------------------------------
// This function initializes the experimental setup for eBug.
//-----------------------------------------------------------------------------
void initializeExperiment(int init_type)
{
  RobotFunction   robot_function;
  LightFunction   light_function;
  ScreenUtility   screen_utility;
  HardwareCommand hardware_command;
  LogUtility      log_utility;

  // Initialize the display.
  screen_utility.clearWindow();

  // Initialize the robot information.
  robot_function.initialize(init_type);

  // Initialize the light information.
  light_function.initialize();

  // Connect to HServer if it is for the real-robot experiment.
  if (run_type == HSERVER) hardware_command.initialize(init_type);

  // Increment the trial number.
  trial++;

  // Initialize the log data if it is specified.
  if ((init_type == ALL) && (log_data)) log_utility.initializeLogFile();
}

//-----------------------------------------------------------------------------
// This function displays the command line arguments of eBug.
//-----------------------------------------------------------------------------
void show_usage(void)
{
  fprintf(stderr, "\n");
  fprintf(stderr, "%s usage:\n", program_name);
  fprintf(stderr, "     %s -[rnlNtSHLvdh] -R <robot name>\n", binary_name);
  fprintf(stderr, "       -r   Load the default robot \"%s\".\n", default_robot);
  fprintf(stderr, "       -R <robot name>\n");
  fprintf(stderr, "            Load the specified robot.\n");
  fprintf(stderr, "       -n   Do not load the default robot.\n");
  fprintf(stderr, "       -l   Load the default stimulus \"%s\".\n", default_light);
  fprintf(stderr, "       -N   Do not load the default stimulus.\n");
  fprintf(stderr, "       -t   Show the robot trails.\n");
  fprintf(stderr, "       -S   Enable the simulation mode.\n");
  fprintf(stderr, "       -H   Enable the real robot mode.\n");
  fprintf(stderr, "            (Run IPTServer and HServer first.)\n");
  fprintf(stderr, "       -L   Log the robot data to \"%s\".\n", log_filename);
  fprintf(stderr, "       -v   Show the version number.\n");
  fprintf(stderr, "       -d   Enable the debug mode.\n");
  fprintf(stderr, "       -h   Display usage.\n");
  fprintf(stderr, "\n");
}

//-----------------------------------------------------------------------------
// This is the main function.
//-----------------------------------------------------------------------------
int main(int argc, char **argv)
{
  ScreenUtility        screen_utility;
  ListUtility          list_utility;
  TimeUtility          time_utility;

  list_utility.create(&robotnameList);
  list_utility.create(&robotList);
  list_utility.create(&lightList);

  int c;
  char *robotname;

  // Get the aruments in the command line.
  while ((c = getopt(argc, argv, "hvdrltnNLHSR:")) != -1) {
    switch (c) {
    case 'd':
      debug = TRUE;
      break;
    case 'r':
      load_default_robot = TRUE;
      break;
    case 'l':
      load_default_light = TRUE;
      break;
    case 'n':
      load_default_robot = FALSE;
      break;
    case 'N':
      load_default_light = FALSE;
      break;
    case 't':
      show_trails = TRUE;
      break;
    case 'H':
      run_type = HSERVER;
      break;
    case 'S':
      run_type = SIMULATION;
      break;
    case 'L':
      log_data = TRUE;
      break;
    case 'R':
      robotname = strdup(optarg);
      list_utility.push(&robotnameList, (int*)robotname);
      break;
    case 'v':
      fprintf(stderr, "%s version %s\n", program_name, version);
      exit(1);
      break;
    case 'h':
      show_usage();
      exit(1);
      break;
    default:
      break;
    }
  }

  // Load the default robot.
  if (load_default_robot)
    list_utility.push(&robotnameList, (int *)default_robot);

  // Initialize X stuff
  screen_utility.initializeX();

  // get screen size from display structure macro
  screen_num = DefaultScreen(display);

  // size window with enough room for text
  width = (unsigned int) (DisplayWidth(display, screen_num)/1.5);
  height = (unsigned int) (DisplayHeight(display, screen_num)/1.5);

  // Define the fallback resources
  static String fallback_resources[] =
  {
    "eBug.geometry: ",
    "*background: LightGray",
    "*foreground: Black",
    NULL
  };
  char geo[50];
  sprintf( geo, "eBug.geometry: %dx%d+10+10", width, height);
  fallback_resources[0] = geo;

  char window_title[512];
  sprintf(window_title, "%s v%s: %s", program_name, version, window_sub_title);
  top_level = XtVaAppInitialize(&app, "eBug", NULL, 0, &argc, argv, fallback_resources,
				XmNtitle, window_title, NULL);
  screen_utility.createWindows(width, height);
  screen_utility.initializeDrawingArea();

  // Realize all widgets
  XtRealizeWidget(top_level);

  // create a pixmap the same size as the drawing area
  drawing_area_pixmap = XCreatePixmap(XtDisplay(drawing_area), RootWindowOfScreen(XtScreen(drawing_area)), 
				      width, height, DefaultDepthOfScreen(XtScreen(drawing_area)));

  // Setup fonts
  XFontStruct *font_info;
  screen_utility.load_font(&font_info);
  screen_utility.getGC(XtWindow(drawing_area),font_info);

  // Setup pixmaps
  plane = XCreatePixmap(XtDisplay(drawing_area),XtWindow(main_window),width,height,DefaultDepthOfScreen(XtScreen(top_level)));
  XSetForeground(XtDisplay(drawing_area), gc, WhitePixel(display,screen_num));
  XFillRectangle(XtDisplay(drawing_area),plane,gc,0,0,width,height);
  XSetForeground(XtDisplay(drawing_area), gc, BlackPixel(display,screen_num));

  // Pop up the hypothesis and orientation need window.
  XtPopup(hypothesis_window_shell, XtGrabNone);
  XtPopup(orientation_need_window_shell, XtGrabNone);

  // Initializing the experiment.
  initializeExperiment(ALL);

  // Start counting the robot age.
  birth_time = time_utility.currentTime();

  // Setup the evet funtions
  screen_utility.getXEvent();

  // Run X loop
  XtAppMainLoop(app);
}


/**********************************************************************
 * $Log: main.cc,v $
 * Revision 1.5  2000/10/28 00:15:30  endo
 * Minor change.
 *
 * Revision 1.4  2000/10/26 07:58:42  endo
 * Minor change.
 *
 * Revision 1.3  2000/10/26 07:43:24  endo
 * Minor change.
 *
 * Revision 1.2  2000/10/25 05:43:02  endo
 * Minor change.
 *
 * Revision 1.1  2000/10/25 04:45:27  endo
 * Initial revision
 *
 * Revision 1.27  2000/08/19 20:46:29  endo
 * Slight changes due to the hardware interface.
 *
 * Revision 1.26  2000/08/13 09:01:46  endo
 * getCurrentLightInfo added.
 *
 * Revision 1.25  2000/08/10 06:15:54  endo
 * driveRobot, stopRobot etc.
 *
 * Revision 1.24  2000/08/08 09:10:17  endo
 * HServer implemented.
 *
 * Revision 1.23  2000/08/07 21:01:29  endo
 * Reset feature added.
 *
 * Revision 1.22  2000/08/07 01:41:51  endo
 * Orientation Need Window created.
 *
 * Revision 1.21  2000/08/07 00:53:18  endo
 * simbugMainLoop moved from ScreenUtility::simulateRobots.
 *
 * Revision 1.20  2000/08/06 07:27:43  endo
 * Two light types implemented successfully.
 *
 * Revision 1.19  2000/07/31 05:31:54  endo
 * Hypothesis feature improved. Orientation Need feature added.
 *
 * Revision 1.18  2000/07/29 22:17:27  endo
 * Reward stuff implemented.
 *
 * Revision 1.17  2000/05/02 14:03:32  endo
 * Minor cleaning up.
 *
 * Revision 1.16  2000/05/02 13:38:27  endo
 * Minor cleaning up.
 *
 * Revision 1.15  2000/05/02 13:10:19  endo
 * Now I can display it on a linux machine. The 5th input for
 * XCreatePixmap was changed from SCREEN_DEPTH to
 * DefaultDepthOfScreen(XtScreen(top_level)).
 *
 * Revision 1.14  2000/04/29 23:13:45  endo
 * drawTrail worked.
 *
 * Revision 1.13  2000/04/29 07:32:13  endo
 * Hypothesis Window added.
 *
 * Revision 1.12  2000/04/27 21:27:26  endo
 * buttons added.
 *
 * Revision 1.11  2000/03/26 10:18:00  endo
 * drawLight and drawSensors also work fine.
 *
 * Revision 1.10  2000/03/26 08:34:35  endo
 * Finally, the simulation works as it supposed to be.
 *
 * Revision 1.9  2000/03/26 05:16:35  endo
 * Got rid of Start. Run/Pause instead.
 *
 * Revision 1.8  2000/03/13 06:36:32  endo
 * Start and Pause works. Getting right input (e.g., key
 * board char, mouse etc.) is still having a problem.
 *
 * Revision 1.7  2000/03/12 09:09:38  endo
 * got rid of pthread, since it doesn't do any
 * good.
 *
 * Revision 1.6  2000/03/12 01:35:37  endo
 * kb_input moved to screen.cc
 *
 * Revision 1.5  2000/03/05 05:58:17  endo
 * pthread implemented. kebyoard vs running robot is
 * the issue to be solved.
 *
 * Revision 1.4  2000/03/05 00:23:59  endo
 * robot started moving.
 *
 * Revision 1.3  2000/03/03 08:27:21  endo
 * It doesn't move, but the robot started showing up
 * in my Motif Window.
 *
 * Revision 1.2  2000/03/03 04:59:34  endo
 * lots of Xm stuff added.
 *
 * Revision 1.1  2000/03/03 01:20:53  endo
 * Initial revision
 *
 **********************************************************************/
