/**********************************************************************
 **                                                                  **
 **                            robot.cc                              **
 **                                                                  **
 **                                                                  **
 **  This file contains the functions for RobotFunction class,       **
 **  which deals with the robots.                                    **
 **                                                                  **
 **  Written by:  Yoichiro Endo (Mobile Robot Lab, Georgia Tech)     **
 **                                                                  **
 **  Copyright 2000 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: robot.cc,v 1.10 2000/10/25 04:45:27 endo Exp $ */

#include "ebug.h"

//-----------------------------------------------------------------------------
// RobotFunction is a class which is a collection of functions that deal with
// robot relatelated issue.
//-----------------------------------------------------------------------------
RobotFunction::RobotFunction()
{
}

//-----------------------------------------------------------------------------
// This function creates the robot list from the name of the robots that are
// specifed to be loaded.
//-----------------------------------------------------------------------------
void RobotFunction::initialize(int init_type)
{
  ListUtility list_utility;

  t_node *tmp_robotList;
  list_utility.create(&tmp_robotList);

  t_node *prn = robotnameList;
  while (prn->next != NULL)
  {
    char *robotname = (char *)prn->data;
    t_robot *tmpRobot = (t_robot *)malloc(sizeof(t_robot));
    load(tmpRobot, robotname);

    // If it is a partial initialization, copy the dynamic parameters from
    // the existing robot.
    if (init_type == PARTIAL)
    {
      t_node *pr = robotList;
      while (pr->next != NULL)
      {
	t_robot *robot = (t_robot *)robotList->data;
	if (tmpRobot->id == robot->id)
	{
	  for (int light_type = 0; light_type < NUM_TYPES; light_type++)
	  {
	    tmpRobot->hypothesis[light_type] = robot->hypothesis[light_type];
	    tmpRobot->orientation_need[light_type] = robot->orientation_need[light_type];
	  }
	  break;
	}
	pr = pr->next;
      }
    }
    list_utility.push(&tmp_robotList, (int*)tmpRobot);

    if (debug) fprintf(stderr, "The robot \"%s\" loaded.\n", robotname);

    prn = prn->next;
  }
  if (robotList->data != NULL) list_utility.clearList(&robotList);

  robotList = tmp_robotList;
}

//-----------------------------------------------------------------------------
// This function updates the robot position based on its forward velocity
// (progression) and angular velocity (angvel).
//-----------------------------------------------------------------------------
void RobotFunction::moveRobot(t_robot *robot)
{
  Calculate calculate;

  robot->last_theta = robot->theta;
  robot->last_x = robot->x;
  robot->last_y = robot->y;
  robot->theta += robot->angvel;

  // Note: robot.theta is zero when it is facing down.
  double angle = robot->theta - 90;
  calculate.convertTheta(&angle);
  robot->x += robot->progression*cos(angle*M_PI/180.0);
  robot->y += robot->progression*sin(angle*M_PI/180.0);
}

//-----------------------------------------------------------------------------
// This function will move the robot after its speed is calculated.
//-----------------------------------------------------------------------------
void RobotFunction::driveRobot(void)
{
  Calculate     calculate;
  
  t_node *p = robotList;
  if (p->data == NULL) return;

  do {
    t_robot *robot = (t_robot *)p->data;

    double force, angle, angvel;

    switch (robot->type) {

    case  TOLMAN_SOWBUG:
      calculate.AngularSpeed(robot);
      calculate.ForwardSpeed(robot);
      moveRobot(robot);
      break;

    case BRAITENBERG_VEHICLE:
      calculate.BVsensorValues();
      calculate.BVmotorForce(robot);
      calculate.BVtotalForce(robot, &force, &angle);
      calculate.BVtorque(robot, &angvel);
      robot->theta += angvel;
      BVmoveRobot(&(robot->x),&(robot->y),angle,force);
      break;
    }

    p = p->next;
  } while (p->next != NULL);
}

//-----------------------------------------------------------------------------
// This function checks whether the robot is moving or not.
//-----------------------------------------------------------------------------
int RobotFunction::isMoving(void)
{
  t_node *p = robotList;
  if (p->data == NULL) return FALSE;

  do {
    t_robot *robot = (t_robot *)p->data;

    if ((robot->x != robot->last_x) || (robot->y != robot->last_y) || (robot->theta != robot->last_theta))
      return TRUE;

    p = p->next;
  } while (p->next != NULL);

  return FALSE;
}

//-----------------------------------------------------------------------------
// This function loads the robot configuration from the file.
//-----------------------------------------------------------------------------
void RobotFunction::load(struct t_robot *robot, char *filename)
{
  ListUtility list_utility;
  FILE        *fp;

  char   buf1[100], buf2[100];
  double val;

  struct t_motor *motor;
  struct t_line *line;
  struct t_sensor *sensor;
  struct t_vector *vector;
  struct t_photo_sensor *photo_sensor;
  struct t_neuron *neuron;

  // Create the lists for the components of the robot.
  list_utility.create(&(robot->lineList));
  list_utility.create(&(robot->motorList));
  list_utility.create(&(robot->sensorList));
  list_utility.create(&(robot->vectorList[ORIENTATION][TYPE_A]));
  list_utility.create(&(robot->vectorList[ORIENTATION][TYPE_B]));
  list_utility.create(&(robot->vectorList[PROGRESSION][TYPE_A]));
  list_utility.create(&(robot->vectorList[PROGRESSION][TYPE_B]));
  list_utility.create(&(robot->photo_sensorList));
  list_utility.create(&(robot->neuronList));
  list_utility.create(&(robot->trailList));

  // Assign the default values first.
  robot->x = DEFAULT_X;
  robot->y = DEFAULT_Y;
  robot->last_x = DEFAULT_X;
  robot->last_y = DEFAULT_Y;
  robot->theta = DEFAULT_THETA;
  robot->last_theta = DEFAULT_THETA;
  robot->trailx = DEFAULT_TRAILX;
  robot->traily = DEFAULT_TRAILY;
  robot->angvel = 0;
  robot->speed = 0;
  robot->progression = 0;
  robot->last_angvel = 0;
  robot->vte = 0;
  robot->bug_height = 25;
  robot->bug_width = 25;
  robot->last_seen_light_type = NO_TYPE;
  robot->hypothesis[TYPE_A] = HYPOTHESIS_DEFAULT_VALUE;
  robot->hypothesis[TYPE_B] = HYPOTHESIS_DEFAULT_VALUE;
  robot->orientation_need[TYPE_A] = ORIENTATION_NEED_DEFAULT_VALUE;
  robot->orientation_need[TYPE_B] = ORIENTATION_NEED_DEFAULT_VALUE;

  // Open the configuration file.
  fp = fopen(filename,"r");
  if (fp == NULL)
  {
    fprintf(stderr, "Error: \"%s\" cannot be opened.\n", filename);
    exit(2);
  }

  // Scan throght the file.
  do {
    fscanf(fp,"%s %s",buf1,buf2);
    if (feof(fp))
    {
      fclose(fp);
      return;
    }
    // Change the string to be lower case.
    for (int i=0; i<(int)strlen(buf1); i++) buf1[i] = tolower(buf1[i]);

    if (strlen(buf2)>0)            sscanf(buf2,"%lf",&val);

    // Get the robot name.
    if (!strcmp(buf1,"name"))      strcpy(robot->name,buf2);

    // Get the robot id.
    if (!strcmp(buf1,"id"))        robot->id = (int)val;

    // Get the robot type (Tolman's Schematic Sowbug or Braitenberg's Vehicle). 
    if (!strcmp(buf1,"type"))      robot->type = (int)val;

    // Get the robot position.
    if (!strcmp(buf1,"x"))
    {
      robot->x = val;
      robot->last_x = val;
    }
    if (!strcmp(buf1,"y"))
    {
      robot->y = val;
      robot->last_y = val;
    }

    // Get the robot heading.
    if (!strcmp(buf1,"theta"))
    {
      robot->theta = val;
      robot->last_theta = val;
    }

    // Get the robot width.
    if (!strcmp(buf1,"width"))      robot->width = (int)val;

    // Get the robot mass.
    if (!strcmp(buf1,"mass"))       robot->mass  = val;

    // Get the robot angular velocity.
    if (!strcmp(buf1,"angvel"))     robot->angvel = val;

    // Get the robot speed.
    if (!strcmp(buf1,"speed"))      robot->speed = val;

    // Get the robot center of mass x.
    if (!strcmp(buf1,"cmx"))        robot->cmx = (int)val;
    if (!strcmp(buf1,"cmy"))        robot->cmy = (int)val;

    // Get the trail position to be drawn.
    if (!strcmp(buf1,"trailx"))     robot->trailx = val;
    if (!strcmp(buf1,"traily"))     robot->traily = val;

    // Get the robot's outline.
    if (!strcmp(buf1,"lines"))
    {
      for(int i=0;i<val;i++)
      {
	line = (t_line *)malloc(sizeof(t_line));
	fscanf(fp,"%d %d %d %d", &(line->x1), &(line->y1), &(line->x2), &(line->y2));
	list_utility.push(&(robot->lineList), (int*)line);
      }
    }

    // Get the robot (for Schematic Sowbug) height and width.
    if (!strcmp(buf1,"bug-height")) robot->bug_height = (unsigned int)val;
    if (!strcmp(buf1,"bug-width"))  robot->bug_width = (unsigned int)val;

    // Get the vector related information (for Schematic Sowbug).
    if (!strcmp(buf1,"vectors"))
    {
      for(int i=0;i<val;i++)
      {
	vector = (t_vector *)malloc(sizeof(t_vector));
	int light_type;
	fscanf(fp,"%d %d %d %d %d %d",
	       &(vector->id),
	       &(vector->disp_x),
	       &(vector->disp_y),
	       &(vector->side),
	       &(vector->type),
	       &light_type);
	list_utility.push(&(robot->vectorList[vector->type][light_type]), (int*)vector);
      }
    }

    // Get the photo-sensor related information (for Schematic Sowbug).
    if (!strcmp(buf1,"photo-sensors"))
    {
      for(int i=0;i<val;i++)
      {
	photo_sensor = (t_photo_sensor *)malloc(sizeof(t_photo_sensor));
	photo_sensor->value[TYPE_A] = 0;
	photo_sensor->value[TYPE_B] = 0;
	fscanf(fp,"%d %d %d %lf %lf %lf %d",
	       &(photo_sensor->id),
	       &(photo_sensor->sensor_x),
	       &(photo_sensor->sensor_y),
               &(photo_sensor->angle),
	       &(photo_sensor->fov),
	       &(photo_sensor->sensitivity),
	       &(photo_sensor->side));
	list_utility.push(&(robot->photo_sensorList),(int*)photo_sensor);
      }
    }

    // Get the motor related information (for Braitenberg Vehicle).
    if (!strcmp(buf1,"motors"))
    {
      for(int i=0;i<val;i++)
      {
	motor = (t_motor *)malloc(sizeof(struct t_motor));
	fscanf(fp,"%d %d %d %lf %lf %lf",
	       &(motor->id),
	       &(motor->x),
	       &(motor->y),
	       &(motor->angle),
	       &(motor->strength),
	       &(motor->maxForce));
	list_utility.push(&(robot->motorList), (int*)motor);
      }
    }

    // Get the robto maximum torque value (for Braitenberg Vehicle).
    if (!strcmp(buf1,"maxtorque"))  robot->maxTorque = val;

    // Get the sensor related information (for Braitenberg Vehicle).
    if (!strcmp(buf1,"sensors"))
    {
      for(int i=0;i<val;i++)
      {
	sensor = (t_sensor *)malloc(sizeof(struct t_sensor));
	fscanf(fp,"%d %d %d %lf %lf %lf",
	       &(sensor->id),
	       &(sensor->x),
	       &(sensor->y),
               &(sensor->angle),
	       &(sensor->fov),
	       &(sensor->sensitivity));
	list_utility.push(&(robot->sensorList),(int*)sensor);
      }
    }

    // Get the neuron related information (for Braitenberg Vehicle).
    if (!strcmp(buf1,"neurons")) {
      for(int i=0;i<val;i++){
	neuron = (t_neuron *)malloc(sizeof(t_neuron));
	fscanf(fp,"%d %d %d",&(neuron->sensorId),&(neuron->motorId),&(neuron->neuronType));
	list_utility.push(&(robot->neuronList),(int*)neuron);
      }
    }
  } while(!feof(fp));     
  fclose(fp);
}

//-----------------------------------------------------------------------------
// This function updates the position of the Braitenberg Vehicle type robot.
//-----------------------------------------------------------------------------
void RobotFunction::BVmoveRobot(double *x, double *y, double theta, double distance)
{
  theta = theta * M_PI / 180;
  *x -= distance*sin(theta);
  *y += distance*cos(theta);
}

/**********************************************************************
 * $Log: robot.cc,v $
 * Revision 1.10  2000/10/25 04:45:27  endo
 * Code cleaned up for the distribution.
 *
 * Revision 1.9  2000/08/19 20:46:29  endo
 * Slight changes due to the hardware interface.
 *
 * Revision 1.8  2000/08/07 21:01:06  endo
 * Reset feature added.
 *
 * Revision 1.7  2000/08/06 07:27:23  endo
 * Two light types implemented successfully.
 *
 * Revision 1.6  2000/07/31 05:31:25  endo
 * Hypothesis feature improved. Orientation Need feature added.
 *
 * Revision 1.5  2000/07/30 21:29:27  endo
 * minor clean up.
 *
 * Revision 1.4  2000/04/30 08:02:41  endo
 * photo_sensor implemented.
 *
 * Revision 1.3  2000/04/29 23:13:21  endo
 * drawTrail worked.
 *
 * Revision 1.2  2000/04/29 09:18:39  endo
 * Progression Vector working.
 *
 * Revision 1.1  2000/03/27 00:49:21  endo
 * Initial revision
 *
 *
 **********************************************************************/
