/**********************************************************************
 **                                                                  **
 **                            screen.cc                             **
 **                                                                  **
 **                                                                  **
 **  Thus file contains the functions for Calculate class, which     **
 **  deals with various calculations in eBug.                        **
 **                                                                  **
 **  Written by:  Yoichiro Endo (Mobile Robot Lab, Georgia Tech)     **
 **                                                                  **
 **  Copyright 2000 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: screen.cc,v 1.42 2000/10/25 04:45:27 endo Exp $ */

#include "ebug.h"

#include "bitmaps/run.bit"
#include "bitmaps/pause.bit"
#include "bitmaps/reset.bit"
#include "bitmaps/quit.bit"
#include "bitmaps/loadlights.bit"

Pixmap pm_run_btn;
Pixmap pm_pause_btn;
Pixmap pm_reset_btn;
Pixmap pm_quit_btn;
Pixmap pm_loadlights_btn;

#define ARROWHEAD_SIZE           5
#define WINDOW2_ROBOT_X        600
#define WINDOW2_ROBOT_Y        400
#define WINDOW2_W              100
#define WINDOW2_H              100
#define OVDC                     1  // O-VECTOR DISPLAY CONSTANT
#define OVD_MAX                 85  // O-VECTOR DISPLAY MAX
#define PVDC                     1  // P-VECTOR DISPLAY CONSTANT
#define PVD_MAX                 85  // P-VECTOR DISPLAY MAX
#define IGC                      2  // INTERNAL GRAPH CONSTANT
#define IGD_MAX                 10  // INTERNAL GRAPH DISPLAY MAX
#define TL_MIN                   2  // TRAIL LENGTH MINIMUM

//-----------------------------------------------------------------------------
// ScreenUtility is a class which is a collection of functions that deal with
// the window display related issue.
//-----------------------------------------------------------------------------
ScreenUtility::ScreenUtility()
{
}

//-----------------------------------------------------------------------------
// This function creates the main window.
//-----------------------------------------------------------------------------
Widget ScreenUtility::MainWindow(Widget parent)
{
  Widget main_window = XtVaCreateManagedWidget("eBug", xmMainWindowWidgetClass,
                                   parent, NULL);

  Pixel fg, bg;

  XtVaGetValues(main_window,
		XmNforeground, &fg,
		XmNbackground, &bg,
		NULL);

  buildBitMaps(fg, bg);

  return main_window;
}

//-----------------------------------------------------------------------------
// Thus function creates the menu bar.
//-----------------------------------------------------------------------------
Widget ScreenUtility::MenuBar(Widget parent)
{
      XmString file      = XmStringCreateSimple("File");

      Widget menu_bar = XmVaCreateSimpleMenuBar(
	 parent, "menu-bar",
	 NULL);

      XmStringFree(file);

      return menu_bar;
}

//-----------------------------------------------------------------------------
// This function creates the work area.
//-----------------------------------------------------------------------------
Widget ScreenUtility::WorkArea(Widget parent)
{
      Widget work_area =  XtVaCreateManagedWidget(
                         "work-area", xmFormWidgetClass, parent, 
                         XmNresizePolicy,          XmNONE,
                         NULL);

      return work_area;
}

//-----------------------------------------------------------------------------
// This function creates the command bar.
//-----------------------------------------------------------------------------
Widget ScreenUtility::CommandBar(Widget parent)
{
      Widget command_bar =  XtVaCreateManagedWidget(
                                 "command-bar", xmFormWidgetClass, parent,
                                 XmNtopAttachment, XmATTACH_FORM,
                                 XmNleftAttachment, XmATTACH_FORM,
                                 XmNrightAttachment, XmATTACH_FORM,
                                 NULL);
      
      return command_bar;
}

//-----------------------------------------------------------------------------
// This function creates the "Load Lights" button.
//-----------------------------------------------------------------------------
Widget ScreenUtility::LoadLightsButton(Widget parent)
{

  Pixel fg, bg;

  XtVaGetValues(main_window,
		XmNforeground, &fg,
		XmNbackground, &bg,
		NULL);

  load_lights_button = XtVaCreateManagedWidget
    ("load-lights-button",
     xmPushButtonWidgetClass, parent,
     XmNlabelType, XmPIXMAP,
     XmNlabelPixmap, pm_loadlights_btn,
     XmNbackground, bg,
     XmNtopAttachment,     XmATTACH_FORM,
     XmNbottomAttachment,  XmATTACH_FORM,
     XmNleftAttachment,    XmATTACH_FORM,
     XmNleftOffset,        1,
     NULL);

  XtAddCallback(load_lights_button, XmNactivateCallback,
		(XtCallbackProc) pushLoadLightsButton_cb, NULL);

  return load_lights_button;
}

//-----------------------------------------------------------------------------
// This function creates the "Run/Pause" button.
//-----------------------------------------------------------------------------
Widget ScreenUtility::RunPauseButton(Widget parent, Widget left_widget)
{

  Pixel fg, bg;

  XtVaGetValues(main_window,
		XmNforeground, &fg,
		XmNbackground, &bg,
		NULL);

  run_pause_button = XtVaCreateManagedWidget
    ("run-pause-button",
     xmPushButtonWidgetClass, parent,
     XmNlabelType, XmPIXMAP,
     XmNlabelPixmap, pm_run_btn,
     XmNbackground, bg,
     XmNtopAttachment,     XmATTACH_FORM,
     XmNbottomAttachment,  XmATTACH_FORM,
     XmNleftAttachment,    XmATTACH_WIDGET,
     XmNleftWidget,        left_widget,
     NULL);

  XtAddCallback(run_pause_button, XmNactivateCallback,
		(XtCallbackProc) pushRunPauseButton_cb, NULL);

  return run_pause_button;
}

//-----------------------------------------------------------------------------
// This function updates the Pixmap of the "Run/Pause" button based on the
// mode.
//-----------------------------------------------------------------------------
void ScreenUtility::updateRunPauseButton(void)
{
  static int last_mode = !go;

  if (last_mode == go) return;
  else last_mode = go;

  XtUnmanageChild(run_pause_button);

  if (go)
  {
    XtVaSetValues(run_pause_button,
		  XmNlabelPixmap, pm_pause_btn,
		  NULL);
    XtManageChild(run_pause_button);
  }
  else
  {
    XtVaSetValues(run_pause_button,
		  XmNlabelPixmap, pm_run_btn,
		  NULL);
    XtManageChild(run_pause_button);
  }
  
  XtAppAddWorkProc(app, (XtWorkProc)ebugMainLoop, NULL);
}

//-----------------------------------------------------------------------------
// This function creates the "Reset" button.
//-----------------------------------------------------------------------------
Widget ScreenUtility::ResetButton(Widget parent, Widget left_widget)
{
  Pixel fg, bg;

  XtVaGetValues
    (main_window,
     XmNforeground, &fg,
     XmNbackground, &bg,
     NULL);

  XmString str = XmStringCreateSimple("Reset");

  reset_button = XtVaCreateManagedWidget(
     "reset-button", xmPushButtonGadgetClass, parent,
     XmNlabelType, XmPIXMAP,
     XmNlabelPixmap, pm_reset_btn,
     XmNbackground, bg,
     XmNtopAttachment,     XmATTACH_FORM,
     XmNbottomAttachment,  XmATTACH_FORM,
     XmNleftAttachment,    XmATTACH_WIDGET,
     XmNleftWidget,        left_widget,
     NULL);
  XmStringFree(str);

  XtAddCallback(reset_button, XmNactivateCallback, (XtCallbackProc)pushResetButton_cb, NULL);

  return reset_button;
}

//-----------------------------------------------------------------------------
// This function creates the "Quit" button.
//-----------------------------------------------------------------------------
Widget ScreenUtility::QuitButton(Widget parent, Widget left_widget)
{
  Pixel fg, bg;

  XtVaGetValues
    (main_window,
     XmNforeground, &fg,
     XmNbackground, &bg,
     NULL);

  XmString str = XmStringCreateSimple("Quit");

  quit_button = XtVaCreateManagedWidget(
     "quit-button", xmPushButtonGadgetClass, parent,
     XmNlabelType, XmPIXMAP,
     XmNlabelPixmap, pm_quit_btn,
     XmNbackground, bg,
     XmNtopAttachment,     XmATTACH_FORM,
     XmNbottomAttachment,  XmATTACH_FORM,
     XmNleftAttachment,    XmATTACH_WIDGET,
     XmNleftWidget,        left_widget,
     NULL);
  XmStringFree(str);

  XtAddCallback(quit_button, XmNactivateCallback, (XtCallbackProc)pushQuitButton_cb, NULL);

  return quit_button;
}

//-----------------------------------------------------------------------------
// This function creates the scrolled window.
//-----------------------------------------------------------------------------
Widget ScreenUtility::ScrolledWindow(Widget parent, Widget top_widget)
{
      Widget scrolled_window =   XtVaCreateManagedWidget(
                        "scrolled-window", xmScrolledWindowWidgetClass, parent,
                        XmNtopAttachment,               XmATTACH_WIDGET,
                        XmNtopWidget,                   top_widget,
                        XmNleftAttachment,              XmATTACH_FORM,
                        XmNrightAttachment,             XmATTACH_FORM,
                        XmNbottomAttachment,            XmATTACH_FORM,
                        XmNscrollBarDisplayPolicy,      XmAS_NEEDED,
                        XmNscrollingPolicy,             XmAUTOMATIC,
                        XmNvisualPolicy,                XmCONSTANT,
                        NULL);
      
      return scrolled_window;
}

//-----------------------------------------------------------------------------
// This function creates the drawing area.
//-----------------------------------------------------------------------------
Widget ScreenUtility::DrawingArea(Widget parent, unsigned int width, unsigned int height)
{
                        Widget drawing_area =    XtVaCreateWidget(
                        "drawing-area", xmDrawingAreaWidgetClass, parent,
                        XmNresizePolicy, XmRESIZE_NONE,
                        XmNwidth,        width,
                        XmNheight,       height,
                        NULL);

      return drawing_area;
}

//-----------------------------------------------------------------------------
// This function creates all the widgets for eBug.
//-----------------------------------------------------------------------------
void ScreenUtility::createWindows(unsigned int width, unsigned int height)
{
  // Create widgets.
  main_window =              MainWindow(top_level);
  menu_bar =                 MenuBar(main_window);
  work_area =                WorkArea(main_window);
  command_bar =              CommandBar(work_area);
  load_lights_button =       LoadLightsButton(command_bar);
  run_pause_button =         RunPauseButton(command_bar, load_lights_button);
  reset_button =             ResetButton(command_bar, run_pause_button);
  quit_button =              QuitButton(command_bar, reset_button);
  scrolled_window =          ScrolledWindow(work_area, command_bar);
  drawing_area =             DrawingArea(scrolled_window, width, height);
  hypothesis_window =        HypothesisWindow(drawing_area);
  orientation_need_window =  OrientationNeedWindow(drawing_area);

  // Manage them.
  XtManageChild(main_window);
  XtManageChild(menu_bar);
  XtManageChild(work_area);
  XtManageChild(command_bar);
  XtManageChild(run_pause_button);
  XtManageChild(scrolled_window);
  XtManageChild(drawing_area);
  XtManageChild(hypothesis_window);
  XtManageChild(orientation_need_window);

  // Attach the menubar and the drawing_area area to the main window
  XmMainWindowSetAreas(main_window, menu_bar, NULL, NULL, NULL, work_area);
}

//-----------------------------------------------------------------------------
// This function initializes the X related issue.
//-----------------------------------------------------------------------------
void ScreenUtility::initializeX(void)
{
  XSizeHints *size_hints;
  if (!(size_hints = XAllocSizeHints()))
  {
    fprintf(stderr, "Error: failure allocating memory\n");
    exit(0);
  }

  XWMHints *wm_hints;
  if (!(wm_hints = XAllocWMHints()))
  {
    fprintf(stderr, "Error: failure allocating memory\n");
    exit(0);
  }

  XClassHint *class_hints;
  if (!(class_hints = XAllocClassHint()))
  {
    fprintf(stderr, "Error: failure allocating memory\n");
    exit(0);
  }

  /* connect to X server */
  char         *display_name = NULL;
  if ( (display = XOpenDisplay(display_name)) == NULL )
  {
    fprintf( stderr, "Error: cannot connect to X server %s\n", XDisplayName(display_name));
    exit( -1 );
  }

}

//-----------------------------------------------------------------------------
// This function initializes the drawing area.
//-----------------------------------------------------------------------------
void ScreenUtility::initializeDrawingArea(void)
{

  int fg, bg, impact_pixel, whitepixel;

  XGCValues xgcv;
  Display *dpy = XtDisplay(drawing_area);
  int scr = DefaultScreen(dpy);
  Visual *vis = DefaultVisual(dpy,scr);

  // set the default color for the impact circle
  impact_pixel = BlackPixel(dpy, scr);
   
  // figure out background and foreground for various types of systems
  if (vis->map_entries > 2) 
  {
    // It's a color display
    fg = WhitePixel(dpy, scr);
    whitepixel = WhitePixel(dpy, scr);
    bg = BlackPixel(dpy, scr);

    XtVaSetValues
      (drawing_area,
       XmNborderWidth, 1,
       XmNborderColor, BlackPixel(dpy,scr),
       XmNbackground, whitepixel,
       NULL);
  }
   
  // Create color pixels.
  XColor color, ignore;
  Colormap cmap = DefaultColormap(dpy, DefaultScreen(display));
  if (XAllocNamedColor(dpy, cmap, "lawn green", &color, &ignore))
    lightcolor_pixel[TYPE_A] = color.pixel;
  if (XAllocNamedColor(dpy, cmap, "deep pink", &color, &ignore))
    lightcolor_pixel[TYPE_B] = color.pixel;
  if (XAllocNamedColor(dpy, cmap, "navy blue", &color, &ignore))
    blue_pixel = color.pixel;
  if (XAllocNamedColor(dpy, cmap, "blue3", &color, &ignore))
    lightblue_pixel = color.pixel;
  if (XAllocNamedColor(dpy, cmap, "firebrick", &color, &ignore))
    red_pixel = color.pixel;
  if (XAllocNamedColor(dpy, cmap, "red2", &color, &ignore))
    pink_pixel = color.pixel;
  if (XAllocNamedColor(dpy, cmap, "green3", &color, &ignore))
    green_pixel = color.pixel;

  vector_pixel[ORIENTATION][TYPE_A] = blue_pixel;
  vector_pixel[ORIENTATION][TYPE_B] = lightblue_pixel;
  vector_pixel[PROGRESSION][TYPE_A] = red_pixel;
  vector_pixel[PROGRESSION][TYPE_B] = pink_pixel;
  graph_pixel[ORIENTATION][TYPE_A] = blue_pixel;
  graph_pixel[ORIENTATION][TYPE_B] = lightblue_pixel;
  graph_pixel[PROGRESSION][TYPE_A] = red_pixel;
  graph_pixel[PROGRESSION][TYPE_B] = pink_pixel;

  // define a GC to xor
  xgcv.foreground = bg;
  xgcv.background = fg;
  xorGC = XtGetGC(drawing_area, GCForeground | GCBackground, &xgcv);
  XSetFunction(XtDisplay(drawing_area), xorGC, GXxor);

  //xgcv.foreground = impact_pixel;
  xgcv.foreground = red_pixel;
  //xgcv.background = fg;
  xgcv.background = red_pixel;
  red_xorGC = XtGetGC(drawing_area, GCForeground | GCBackground, &xgcv);
  XSetFunction(XtDisplay(drawing_area), red_xorGC, GXxor);

  // define a normal drawing GC
  xgcv.foreground = fg;
  xgcv.background = bg;
  theGC = XtGetGC(drawing_area, GCForeground | GCBackground, &xgcv);
}

//-----------------------------------------------------------------------------
// This function handles the events in the windows. 
//-----------------------------------------------------------------------------
void ScreenUtility::getXEvent(void)
{

  XSelectInput(XtDisplay(drawing_area), XtWindow(drawing_area), event_mask);
  XtAddCallback(drawing_area, XmNinputCallback, (XtCallbackProc)kb_input_cb, NULL);

  while(1)
  {
    XtAppNextEvent(app, &event);
    XtDispatchEvent (&event);

    if (XCheckMaskEvent(XtDisplay(drawing_area), event_mask, &event))
      XtDispatchEvent (&event);
    else
      XtAppAddWorkProc(app, (XtWorkProc)ebugMainLoop, NULL);
  }
}

//-----------------------------------------------------------------------------
// This function deals with creating the graphic contexes in the windows.
//-----------------------------------------------------------------------------
void ScreenUtility::getGC(Window win, XFontStruct *font_info)
{
  XGCValues    values;
    
  unsigned long valuemask = 0; // ignore XGCvalues and use defaults
  unsigned int line_width = 2;
  int line_style = LineSolid;
  int cap_style = CapRound;
  int join_style = JoinRound;

  // Regular gc
  gc = XCreateGC(XtDisplay(drawing_area), win, valuemask, &values);
  XSetFont(XtDisplay(drawing_area), gc, font_info->fid);
  XSetForeground(XtDisplay(drawing_area), gc, BlackPixel(XtDisplay(drawing_area), DefaultScreen(display)));
  XSetLineAttributes(XtDisplay(drawing_area), gc, line_width, line_style, cap_style, join_style);

  // Blue line gc.
  values.foreground = blue_pixel;
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  blueline1gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			 (GCForeground | GCBackground), &values);
  line_width = 1;
  XSetLineAttributes(XtDisplay(drawing_area), blueline1gc, line_width, line_style, cap_style, join_style);

  blueline3gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			 (GCForeground | GCBackground), &values);
  line_width = 3;
  XSetLineAttributes(XtDisplay(drawing_area), blueline3gc, line_width, line_style, cap_style, join_style);

  values.foreground = lightblue_pixel;
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  lightblueline1gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			 (GCForeground | GCBackground), &values);
  line_width = 1;
  XSetLineAttributes(XtDisplay(drawing_area), lightblueline1gc, line_width, line_style, cap_style, join_style);

  lightblueline3gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			 (GCForeground | GCBackground), &values);
  line_width = 3;
  XSetLineAttributes(XtDisplay(drawing_area), lightblueline3gc, line_width, line_style, cap_style, join_style);

  // Red line gc.
  values.foreground = red_pixel;
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  redline1gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			(GCForeground | GCBackground), &values);
  line_width = 1;
  XSetLineAttributes(XtDisplay(drawing_area), redline1gc, line_width, line_style, cap_style, join_style);

  redline3gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			(GCForeground | GCBackground), &values);
  line_width = 3;
  XSetLineAttributes(XtDisplay(drawing_area), redline3gc, line_width, line_style, cap_style, join_style);

  // Pink line gc.
  values.foreground = pink_pixel;
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  pinkline1gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			(GCForeground | GCBackground), &values);
  line_width = 1;
  XSetLineAttributes(XtDisplay(drawing_area), pinkline1gc, line_width, line_style, cap_style, join_style);

  pinkline3gc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
			(GCForeground | GCBackground), &values);
  line_width = 3;
  XSetLineAttributes(XtDisplay(drawing_area), pinkline3gc, line_width, line_style, cap_style, join_style);

  // Trail gc.
  values.foreground = BlackPixel(XtDisplay(drawing_area), DefaultScreen(display));
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  trailgc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
		      (GCForeground | GCBackground), &values);
  line_width = 1;
  XSetLineAttributes(XtDisplay(drawing_area), trailgc, line_width, line_style, cap_style, join_style);

  // Light gc.
  values.foreground = lightcolor_pixel[TYPE_A];
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  lightgc[TYPE_A] = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
		      (GCForeground | GCBackground), &values);
  line_width = 2;
  XSetLineAttributes(XtDisplay(drawing_area), lightgc[TYPE_A], line_width, line_style, cap_style, join_style);

  values.foreground = lightcolor_pixel[TYPE_B];
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  lightgc[TYPE_B] = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
		      (GCForeground | GCBackground), &values);
  line_width = 2;
  XSetLineAttributes(XtDisplay(drawing_area), lightgc[TYPE_B], line_width, line_style, cap_style, join_style);

  // Sensor gc.
  values.foreground = green_pixel;
  values.background = WhitePixel(XtDisplay(drawing_area), DefaultScreen(display));
  sensorgc = XCreateGC(XtDisplay(drawing_area), RootWindow(XtDisplay(drawing_area), DefaultScreen(display)),
		      (GCForeground | GCBackground), &values);
  line_width = 1;
  XSetLineAttributes(XtDisplay(drawing_area), sensorgc, line_width, line_style, cap_style, join_style);

  // Vector gc
  vectorgc[ORIENTATION][TYPE_A] = blueline3gc;
  vectorgc[ORIENTATION][TYPE_B] = lightblueline3gc;
  vectorgc[PROGRESSION][TYPE_A] = redline3gc;
  vectorgc[PROGRESSION][TYPE_B] = pinkline3gc;

  // Graph gc
  graphgc[ORIENTATION][TYPE_A] = blueline1gc;
  graphgc[ORIENTATION][TYPE_B] = lightblueline1gc;
  graphgc[PROGRESSION][TYPE_A] = redline1gc;
  graphgc[PROGRESSION][TYPE_B] = pinkline1gc;
}

//-----------------------------------------------------------------------------
// This function loads the font to be used in the windows.
//-----------------------------------------------------------------------------
void ScreenUtility::load_font(XFontStruct **font_info)
{
  char *fontname = "-*-fixed-bold-r-*-*-14-*";
  
  /* Load font and get font information structure. */
  if ((*font_info = XLoadQueryFont(display,fontname)) == NULL)
  {
    fprintf(stderr, "Error: Cannot open %s font\n", fontname);
    exit( -1 );
  }
}

//-----------------------------------------------------------------------------
// This function redraws the drawing area.
//-----------------------------------------------------------------------------
void ScreenUtility::redrawWindow(void)
{

  if (robotList->data != NULL) drawRobots(DRAW, PRIMARY);
  if (lightList->data != NULL) drawLights(DRAW);

  int x = DisplayWidth(XtDisplay(drawing_area), screen_num);
  int y = DisplayHeight(XtDisplay(drawing_area), screen_num);

  XSetForeground(XtDisplay(drawing_area), gc, BlackPixel(XtDisplay(drawing_area), DefaultScreen(display)));
  XCopyArea(XtDisplay(drawing_area), drawing_area_pixmap, XtWindow(drawing_area), gc, 0, 0, x, y, 0, 0);

}

//-----------------------------------------------------------------------------
// This function clears the drawing area.
//-----------------------------------------------------------------------------
void ScreenUtility::clearWindow(void)
{
  if (robotList->data != NULL) drawRobots(ERASE, PRIMARY);
  if (lightList->data != NULL) drawLights(ERASE);

  int x = DisplayWidth(XtDisplay(drawing_area), screen_num);
  int y = DisplayHeight(XtDisplay(drawing_area), screen_num);

  XSetForeground(XtDisplay(drawing_area), gc, WhitePixel(XtDisplay(drawing_area), DefaultScreen(display)));
  XCopyArea(XtDisplay(drawing_area), drawing_area_pixmap, XtWindow(drawing_area), gc, 0, 0, x, y, 0, 0);

}

//-----------------------------------------------------------------------------
// This function clears and redraws the drawing area.
//-----------------------------------------------------------------------------
void ScreenUtility::refreshWindow(void)
{
  clearWindow();
  redrawWindow();
}

//-----------------------------------------------------------------------------
// This function creates a window for Hypothesis values.
//-----------------------------------------------------------------------------
Widget ScreenUtility::HypothesisWindow(Widget parent)
{
  char *label;

  /* create the popup */
  hypothesis_window_shell = XtVaCreatePopupShell
    (
     "Hypothesis Window", xmDialogShellWidgetClass, parent, 
     XmNdeleteResponse, XmDESTROY,
     NULL);

  Widget hypothesis_window = XtVaCreateWidget
    (
     "hypothesis-window", xmPanedWindowWidgetClass, hypothesis_window_shell,
     XmNsashHeight,        1,
     XmNsashWidth,         1,
     NULL);

  Widget form1 = XtVaCreateWidget
    (
     "hypothesis-window-form1", xmFormWidgetClass, hypothesis_window,
     NULL);

  label = strdup("Type-A");
  hypothesis_scale[TYPE_A] = XtVaCreateManagedWidget
    (
     "hypothesis-scale1", xmScaleWidgetClass, form1,
     XmNwidth,		        175,
     XmNminimum,		HYPOTHESIS_MIN,
     XmNmaximum,		HYPOTHESIS_MAX,
     XmNvalue,		        HYPOTHESIS_DEFAULT_VALUE,
     XmNdecimalPoints,		0,
     XmNshowValue,		True,
     XmNorientation,		XmVERTICAL,
     XmNtopAttachment,	        XmATTACH_FORM,
     XmNleftAttachment,         XmATTACH_FORM,
     XtVaTypedArg, XmNtitleString, XmRString, label, strlen(label)+1,
     NULL);
  XtAddCallback (hypothesis_scale[TYPE_A], XmNvalueChangedCallback,  updateHypothesisValue_cb, (XtPointer)TYPE_A);
  free(label);

  label = strdup("Type-B");
  hypothesis_scale[TYPE_B] = XtVaCreateManagedWidget
    (
     "hypothesis-scale2", xmScaleWidgetClass, form1,
     XmNwidth,		        175,
     XmNminimum,		HYPOTHESIS_MIN,
     XmNmaximum,		HYPOTHESIS_MAX,
     XmNvalue,		        HYPOTHESIS_DEFAULT_VALUE,
     XmNdecimalPoints,		0,
     XmNshowValue,		True,
     XmNorientation,		XmVERTICAL,
     XmNtopAttachment,	        XmATTACH_FORM,
     XmNleftAttachment,         XmATTACH_WIDGET,
     XmNleftWidget,             hypothesis_scale[0],
     XmNrightAttachment,        XmATTACH_FORM,
     XtVaTypedArg, XmNtitleString, XmRString, label, strlen(label)+1,
     NULL);
  XtAddCallback (hypothesis_scale[TYPE_B], XmNvalueChangedCallback, updateHypothesisValue_cb, (XtPointer)TYPE_B);
  free(label);
  
  XtManageChild(form1);

  return hypothesis_window;

}

//-----------------------------------------------------------------------------
// This function updates the Hypothesis window based on the values.
//-----------------------------------------------------------------------------
void ScreenUtility::updateHypothesisWindow(void)
{
  if (robotList->data == NULL) return;

  t_node *pr = robotList;
  while (pr->next != NULL)
  {
    t_robot *robot = (t_robot *)pr->data;

    for (int light_type = 0; light_type < NUM_TYPES; light_type++)
    {
      int window_value;
      XmScaleGetValue(hypothesis_scale[light_type], &window_value);

      if (window_value != robot->hypothesis[light_type])
	XmScaleSetValue(hypothesis_scale[light_type], (int)robot->hypothesis[light_type]);
    }
    pr = pr->next;
  }
}

//-----------------------------------------------------------------------------
// This function creates a window for the Orientation Need.
//-----------------------------------------------------------------------------
Widget ScreenUtility::OrientationNeedWindow(Widget parent)
{
  char *label;

  /* create the popup */
  orientation_need_window_shell = XtVaCreatePopupShell
    (
     "Orientation Need Window", xmDialogShellWidgetClass, parent, 
     XmNdeleteResponse, XmDESTROY,
     NULL);

  Widget orientation_need_window = XtVaCreateWidget
    (
     "orientation-need-window", xmPanedWindowWidgetClass, orientation_need_window_shell,
     XmNsashHeight,        1,
     XmNsashWidth,         1,
     NULL);

  Widget form1 = XtVaCreateWidget
    (
     "orientation-need-window-form1", xmFormWidgetClass, orientation_need_window,
     NULL);

  label = strdup("Type-A");
  orientation_need_scale[TYPE_A] = XtVaCreateManagedWidget
    (
     "orientation-need-scale1", xmScaleWidgetClass, form1,
     XmNwidth,		        175,
     XmNminimum,		ORIENTATION_NEED_MIN,
     XmNmaximum,		ORIENTATION_NEED_MAX,
     XmNvalue,		        ORIENTATION_NEED_DEFAULT_VALUE,
     XmNdecimalPoints,		0,
     XmNshowValue,		True,
     XmNorientation,		XmVERTICAL,
     XmNtopAttachment,	        XmATTACH_FORM,
     XmNleftAttachment,         XmATTACH_FORM,
     XtVaTypedArg, XmNtitleString, XmRString, label, strlen(label)+1,
     NULL);
  XtAddCallback (orientation_need_scale[TYPE_A], XmNvalueChangedCallback,  updateOrientationNeedValue_cb, (XtPointer)TYPE_A);
  free(label);

  label = strdup("Type-B");
  orientation_need_scale[TYPE_B] = XtVaCreateManagedWidget
    (
     "orientation-need-scale2", xmScaleWidgetClass, form1,
     XmNwidth,		        175,
     XmNminimum,		ORIENTATION_NEED_MIN,
     XmNmaximum,		ORIENTATION_NEED_MAX,
     XmNvalue,		        ORIENTATION_NEED_DEFAULT_VALUE,
     XmNdecimalPoints,		0,
     XmNshowValue,		True,
     XmNorientation,		XmVERTICAL,
     XmNtopAttachment,	        XmATTACH_FORM,
     XmNleftAttachment,         XmATTACH_WIDGET,
     XmNleftWidget,             orientation_need_scale[0],
     XmNrightAttachment,        XmATTACH_FORM,
     XtVaTypedArg, XmNtitleString, XmRString, label, strlen(label)+1,
     NULL);
  XtAddCallback (orientation_need_scale[TYPE_B], XmNvalueChangedCallback, updateOrientationNeedValue_cb, (XtPointer)TYPE_B);
  free(label);
  
  XtManageChild(form1);

  return orientation_need_window;

}

//-----------------------------------------------------------------------------
// This function updates the Orientation Need  window based on the values.
//-----------------------------------------------------------------------------
void ScreenUtility::updateOrientationNeedWindow(void)
{
  if (robotList->data == NULL) return;

  t_node *pr = robotList;
  while (pr->next != NULL)
  {
    t_robot *robot = (t_robot *)pr->data;

    for (int light_type = 0; light_type < NUM_TYPES; light_type++)
    {
      int window_value;
      XmScaleGetValue(orientation_need_scale[light_type], &window_value);

      if (window_value != (int)robot->orientation_need[light_type])
	XmScaleSetValue(orientation_need_scale[light_type], (int)robot->orientation_need[light_type]);
    }
    pr = pr->next;
  }
}

//-----------------------------------------------------------------------------
// This function draws robots on the display.
//-----------------------------------------------------------------------------
void ScreenUtility::drawRobots(int draw, int window_num)
{
  Calculate calculate;
  double x, y, cx, cy, theta;
  static double x1, y1;
  struct t_line *line;
  XSegment segments[MAX_SEGMENTS];
  unsigned int bug_height, bug_width;
  int nsegments;
 
  switch (draw) {
  case DRAW:
    XSetForeground(XtDisplay(drawing_area), gc, BlackPixel(XtDisplay(drawing_area), DefaultScreen(display)));
    break;
  case ERASE:
    XSetForeground(XtDisplay(drawing_area), gc, WhitePixel(XtDisplay(drawing_area), DefaultScreen(display)));
    break;
  }

  // Get the robot list.
  t_node *pr = robotList;
  while (pr->data != NULL)
  {
    // Get an individual robot information.
    t_robot *robot = (t_robot *)pr->data;

    // Get the lines of the robot.
    t_node *pl = robot->lineList;
    if (pl->data == NULL) break;

    int i = 0;

    // Make the lines to be XSegment.
    while (pl->next != NULL)
    {
      line = (t_line *)pl->data;
      segments[i].x1 = line->x1;
      segments[i].y1 = line->y1;
      segments[i].x2 = line->x2;
      segments[i].y2 = line->y2;
      i++;
      pl = pl->next;
    }

    nsegments = i;

    switch (window_num) {
    case PRIMARY:
      x = robot->x;
      y = robot->y;
      break;
    case SECONDARY:
      x = WINDOW2_ROBOT_X;
      y = WINDOW2_ROBOT_Y;
      break;
    }

    cx = robot->cmx;
    cy = robot->cmy;
    theta = robot->theta;
    cx += x;
    cy += y;

    // Robot types: Tolman's Schematic Sowbug, or Britenberg's Vehicle.
    if (robot->type == TOLMAN_SOWBUG)
    {
      bug_height = robot->bug_height;
      bug_width = robot->bug_width;

      double x_front = x;
      double y_front = y - bug_height/2;
      double x_rear = x;
      double y_rear = y + bug_height/2;
      int angle = -(int)robot->theta;

      while (angle >= 180) angle -= 360;
      while (angle < -180) angle +=360;

      calculate.rotatePoint(theta, &x_front, &y_front, x, y);
      calculate.rotatePoint(theta, &x_rear, &y_rear, x, y);
      
      translateSegments(segments,nsegments, x, y);
      rotateSegments(theta,segments,nsegments, cx, cy);
      segments[i].y2 = line->y2;
      convertSegments(segments, nsegments);
      XDrawSegments(XtDisplay(drawing_area), plane, gc, segments, nsegments);

      XDrawArc(XtDisplay(drawing_area), plane, gc, (int)(x_front-bug_width/2.0), height - (int)(y_front+bug_height/2.0),
	       (unsigned int)bug_width, (unsigned int)bug_height, (180-angle)*XDRAWARCUNITS, 180*XDRAWARCUNITS);
      XDrawArc(XtDisplay(drawing_area), plane, gc, (int)(x_rear-bug_width/2.0), height - (int)(y_rear+bug_height/2.0),
	       (unsigned int)bug_width, (unsigned int)bug_height, -angle*XDRAWARCUNITS, 180*XDRAWARCUNITS);

    }
    else if (robot->type == BRAITENBERG_VEHICLE)
    {
      translateSegments(segments,nsegments,x,y);
      rotateSegments(theta,segments,nsegments,cx,cy);
      convertSegments(segments, nsegments);
      segments[i].y2 = line->y2;
      XDrawSegments(XtDisplay(drawing_area), plane, gc, segments, nsegments);
    }

    // Let's draw sensor, vectors, and the Window2 box (if necessary).
    drawSensors(robot, draw, window_num);

    if (robot->type == TOLMAN_SOWBUG)
    {
      drawVectors(robot, ORIENTATION, draw, window_num);
      drawVectors(robot, PROGRESSION, draw, window_num);
      drawInternalGraph(robot, ORIENTATION, draw, window_num);
      drawInternalGraph(robot, PROGRESSION, draw, window_num);
    }

    x1 = cx-robot->width/2;
    y1 = height - cy-robot->width/2;

    // Draw the trail
    if (show_trails) drawTrail(robot, draw);

    if (draw)
    {
      XCopyArea(XtDisplay(drawing_area), plane, XtWindow(drawing_area), gc, (int)x1, (int)y1,
		robot->width, robot->width, (int)x1, (int)y1);
    }
    pr = pr->next;
  }
}

//-----------------------------------------------------------------------------
// This function draws the photo-sensors of the robot.
//-----------------------------------------------------------------------------
void ScreenUtility::drawSensors(t_robot *robot, int draw, int window_num)
{
  if (robot->type != TOLMAN_SOWBUG) return;
    
  Calculate calculate;
  double    x, y, sx,sy;

  switch (draw) {
  case DRAW:
    XSetForeground(XtDisplay(drawing_area), sensorgc, green_pixel);
    break;
  case ERASE:
    XSetForeground(XtDisplay(drawing_area), sensorgc, WhitePixel(display,screen_num));
    break;
  }

  switch (window_num) {
  case PRIMARY:
    x = robot->x;
    y = robot->y;
    break;
  case SECONDARY:
    x = WINDOW2_ROBOT_X;
    y = WINDOW2_ROBOT_Y;
    break;
  }

  t_node *photoList = robot->photo_sensorList;

  if (photoList->data == NULL) return;
  while (photoList->next != NULL)
  {
    t_photo_sensor *photo_sensor = (t_photo_sensor *)photoList->data;
      
    sx = photo_sensor->sensor_x + x;
    sy = photo_sensor->sensor_y + y;
    calculate.rotatePoint(robot->theta, &sx, &sy, robot->cmx+x, robot->cmy+y);

    sx -= photo_sensor->sensitivity/2;
    sy = height - photo_sensor->sensitivity/2 - sy;
    XDrawArc(XtDisplay(drawing_area), plane, sensorgc, (int)sx, (int)sy, (unsigned int)photo_sensor->sensitivity,
	     (unsigned int)photo_sensor->sensitivity,0,360*XDRAWARCUNITS);
    XFillArc(XtDisplay(drawing_area), plane, sensorgc, (int)sx,  (int)sy, (unsigned int)photo_sensor->sensitivity,
	     (unsigned int)photo_sensor->sensitivity, 0, 360*XDRAWARCUNITS);

    photoList = (*photoList).next;
  }
}

//-----------------------------------------------------------------------------
// This function draws the vectors of specified type (Orientaton or
// Progression).
//-----------------------------------------------------------------------------
void ScreenUtility::drawVectors(t_robot *robot, int vector_type, int draw, int window_num)
{
  Calculate calculate;
  double    x, y, vecx, vecy, theta;
  int       i, nsegments, side;
  short int vector_value;

  for (int light_type = 0; light_type < NUM_TYPES; light_type++)
  {
    switch (draw) {
    case DRAW:
      XSetForeground(XtDisplay(drawing_area), vectorgc[vector_type][light_type], vector_pixel[vector_type][light_type]);
      break;
    case ERASE:
      XSetForeground(XtDisplay(drawing_area), vectorgc[vector_type][light_type], WhitePixel(display,screen_num));
      break;
    }

    // Retrieving the vector informatio based of its type.
    t_node *vecList = robot->vectorList[vector_type][light_type];

    if (vecList->data == NULL)
    {
      light_type++;
      continue;
    }

    XSegment segments[MAX_SEGMENTS];

    switch (window_num) {
    case PRIMARY:
      x = robot->x;
      y = robot->y;
      break;
    case SECONDARY:
      x = WINDOW2_ROBOT_X;
      y = WINDOW2_ROBOT_Y;
      break;
    }

    while (vecList->next != NULL)
    {
      t_vector *vector = (t_vector *)vecList->data;
      i = 0;

      segments[i].x1 = 0;
      segments[i].y1 = 0;

      if ((vector_type == ORIENTATION)&&(vector->type == ORIENTATION))
      {
	if (abs(vector_value = (short int)(vector->value*OVDC)) > OVD_MAX)
	  vector_value = OVD_MAX;

	if (vector->side == LEFT) vector_value *= -1;

	segments[i].x2 = vector_value;
	segments[i].y2 = 0;

	side = vector->side;
      }
      else if ((vector_type == PROGRESSION)&&(vector->type == PROGRESSION))
      {
	segments[i].x2 = 0;

	if (abs(vector_value = (short int)vector->value*PVDC) > PVD_MAX)
	  vector_value = PVD_MAX;

	segments[i].y2 = vector_value;

	side = REAR;
      }

      i++;
      
      if (fabs(vector_value) > 0)
	constructArrowHead((XSegment *)segments, side, &i);

      nsegments = i;

      vecx = vector->disp_x + x;
      vecy = vector->disp_y + y;
      calculate.rotatePoint(robot->theta, &vecx, &vecy, robot->cmx+x, robot->cmy+y);

      theta = robot->theta;
      translateSegments(segments,nsegments,vecx,vecy);
      rotateSegments(theta,segments,nsegments,vecx,vecy);
      convertSegments(segments, nsegments);
      XDrawSegments(XtDisplay(drawing_area), plane, vectorgc[vector_type][light_type], segments, nsegments);
      vecList = vecList->next;
    }
  }
}

//-----------------------------------------------------------------------------
// This function draws the distribution graph of the robot.
//-----------------------------------------------------------------------------
void ScreenUtility::drawInternalGraph(t_robot *robot, int vector_type, int draw, int window_num)
{

  double x, y;

  switch (window_num) {
  case PRIMARY:
    x = robot->x;
    y = robot->y;
    break;
  case SECONDARY:
    x = WINDOW2_ROBOT_X;
    y = WINDOW2_ROBOT_Y;
    break;
  }

  XSegment segments[MAX_SEGMENTS];

  for (int light_type = 0; light_type < NUM_TYPES; light_type++)
  {
    switch (draw) {
    case DRAW:
      XSetForeground(XtDisplay(drawing_area), graphgc[vector_type][light_type], graph_pixel[vector_type][light_type]);
      break;
    case ERASE:
      XSetForeground(XtDisplay(drawing_area), graphgc[vector_type][light_type], WhitePixel(display,screen_num));
      break;
    }

    int i = 0;
    double last_x_value, last_y_value;

    // Get the photo-sensor list of the robot.
    t_node *photoList = robot->photo_sensorList;

    while (photoList->next != NULL)
    {
      // Get an individual photo-sensor information.
      t_photo_sensor *photo_sensor = (t_photo_sensor *)photoList->data;

      // Scale the value in order to display with in the robot region.
      double photo_value = photo_sensor->value[light_type]*IGC;
      if (photo_value > IGD_MAX) photo_value = IGD_MAX;

      double x_value = (*photo_sensor).sensor_x + x;
      double y_value;
      switch (vector_type) {
      case ORIENTATION:
	y_value = -photo_value*robot->orientation_need[light_type] + y - robot->bug_height/2;
	break;
      case PROGRESSION:
	y_value = -photo_value*robot->hypothesis[light_type] + y + robot->bug_height/2;
	break;
      }

      if (i == 0)
      {
	last_x_value = x_value;
	last_y_value = y_value;
      }

      // Translate the photo-sensor values to XSegment.
      segments[i].x1 = (int)last_x_value;
      segments[i].y1 = (int)last_y_value;
      segments[i].x2 = (int)x_value;
      segments[i].y2 = (int)y_value;

      last_x_value = x_value;
      last_y_value = y_value;

      i++;
      photoList = photoList->next;
    }
    int nsegments = i;

    // Draw the graph segment after rotating it.
    rotateSegments(robot->theta,segments, nsegments, x, y);
    convertSegments(segments, nsegments);
    XDrawSegments(XtDisplay(drawing_area), plane, graphgc[vector_type][light_type], segments, nsegments);

    // Draw the nodes of the graph.      
    for (i = 0; i<nsegments; i++)
    {
      XDrawArc(XtDisplay(drawing_area), plane, graphgc[vector_type][light_type], segments[i].x2-2, segments[i].y2-2,
	       4, 4, 0, 360*XDRAWARCUNITS);
      XFillArc(XtDisplay(drawing_area), plane, graphgc[vector_type][light_type], segments[i].x2-2, segments[i].y2-2,
	       4, 4, 0, 360*XDRAWARCUNITS);
    }
  }
}

//-----------------------------------------------------------------------------
// This function draws the trail of the robot.
//-----------------------------------------------------------------------------
void ScreenUtility::drawTrail(t_robot *robot, int draw)
{
  // Only TOLMAN_SOWBUG type supports this function.
  if (robot->type != TOLMAN_SOWBUG) return;

  if (draw == DRAW)
  {
    XSetForeground(XtDisplay(drawing_area), trailgc, BlackPixel(XtDisplay(drawing_area), DefaultScreen(display)));

    ListUtility list_utility;
    Calculate   calculate;

    double x = robot->x + robot->trailx;
    double y = robot->y + robot->traily;
    calculate.rotatePoint(robot->theta, &x, &y, robot->cmx+robot->x, robot->cmy+robot->y);

    point last_trailpt; 
    t_node *tl = robot->trailList;

    if (tl->next != NULL)
    {
      last_trailpt.y = (int)tl->data;
      tl = tl->next;
      last_trailpt.x = (int)tl->data;
    }
    else
    {
      last_trailpt.x = (int)x;
      last_trailpt.y = (int)y;
      list_utility.push(&(robot->trailList), (int *)last_trailpt.x);
      list_utility.push(&(robot->trailList), (int *)last_trailpt.y);
    }

    double dx = x - (double)last_trailpt.x;
    double dy = y - (double)last_trailpt.y;
    double diff = sqrt(pow(dx, 2) + pow(dy, 2));

    if (diff >= TL_MIN)
    {
      int x_ = (int)x;
      int y_ = (int)y;
      list_utility.push(&(robot->trailList), (int *)x_);
      list_utility.push(&(robot->trailList), (int *)y_);
    }
  }
  else if (draw == ERASE)
  {
    XSetForeground(XtDisplay(drawing_area), trailgc, WhitePixel(XtDisplay(drawing_area), DefaultScreen(display)));
  }

  int x1, y1, x2, y2;
  XSegment segments[1];

  t_node *tl = robot->trailList;
  if (!(y1 = (int)tl->data)) return;
  tl = tl->next;
  if (!(x1 = (int)tl->data)) return;
  tl = tl->next;

  while (tl->next != NULL)
  {
    if (!(y2 = (int)tl->data)) return;
    tl = tl->next;
    if (!(x2 = (int)tl->data)) return;

    segments[0].x1 = x1;
    segments[0].y1 = y1;
    segments[0].x2 = x2;
    segments[0].y2 = y2;
    convertSegments(segments, 1);
    XDrawSegments(XtDisplay(drawing_area), plane, trailgc, segments, 1);

    x1 = x2;
    y1 = y2;

    tl = tl->next;
  }
}

//-----------------------------------------------------------------------------
// This function construct the arrowhead of the vectorss
//-----------------------------------------------------------------------------
void ScreenUtility::constructArrowHead(XSegment *segments, int side, int *i)
{

  int ah_size = ARROWHEAD_SIZE;

  switch (side) {
  case RIGHT:
    segments[*i].x1 = 0;
    segments[*i].y1 = 0;
    segments[*i].x2 = ah_size;
    segments[*i].y2 = ah_size;
    segments[*i+1].x1 = 0;
    segments[*i+1].y1 = 0;
    segments[*i+1].x2 = ah_size;
    segments[*i+1].y2 = -ah_size;
    segments[*i+2].x1 = 0;
    segments[*i+2].y1 = 0;
    segments[*i+2].x2 = 2*ah_size;
    segments[*i+2].y2 = 0;
    break;

  case LEFT:
    segments[*i].x1 = 0;
    segments[*i].y1 = 0;
    segments[*i].x2 = -ah_size;
    segments[*i].y2 = ah_size;
    segments[*i+1].x1 = 0;
    segments[*i+1].y1 = 0;
    segments[*i+1].x2 = -ah_size;
    segments[*i+1].y2 = -ah_size;
    segments[*i+2].x1 = 0;
    segments[*i+2].y1 = 0;
    segments[*i+2].x2 = -2*ah_size;
    segments[*i+2].y2 = 0;
    break;

  case REAR:
  case FRONT:
    segments[*i].x1 = 0;
    segments[*i].y1 = 0;
    segments[*i].x2 = -ah_size;
    segments[*i].y2 = ah_size;
    segments[*i+1].x1 = 0;
    segments[*i+1].y1 = 0;
    segments[*i+1].x2 = ah_size;
    segments[*i+1].y2 = ah_size;
    segments[*i+2].x1 = 0;
    segments[*i+2].y1 = 0;
    segments[*i+2].x2 = 0;
    segments[*i+2].y2 = 2*ah_size;
    break;
  }

  (*i) += 3;
}

//-----------------------------------------------------------------------------
// This function draws the stimulus (lights) in the display.
//-----------------------------------------------------------------------------
void ScreenUtility::drawLights(int draw)
{
  if (lightList->data == NULL) return;

  t_node *p = lightList;
  int light_type;
  while (p->next != NULL)
  {
    for (light_type = 0; light_type<NUM_TYPES; light_type++)
    {
      switch (draw) {
      case DRAW:
	XSetForeground(XtDisplay(drawing_area), lightgc[light_type], lightcolor_pixel[light_type]);
	break;
      case ERASE:
	XSetForeground(XtDisplay(drawing_area), lightgc[light_type], WhitePixel(display,screen_num));
	break;
      }
    }

    XSegment segments[MAX_SEGMENTS];
    t_light *light = (t_light *)p->data;
    int x = light->x-light->radius;
    int y = height - light->y-light->radius;
    XFillArc(XtDisplay(drawing_area), plane, lightgc[light->type], x, y,
	     light->radius*2, light->radius*2, 0, 360*XDRAWARCUNITS);

    for (light_type = 0; light_type<NUM_TYPES; light_type++)
    {
      switch (draw) {
      case DRAW:
	XSetForeground(XtDisplay(drawing_area), lightgc[light_type], WhitePixel(display,screen_num));
	break;
      case ERASE:
	XSetForeground(XtDisplay(drawing_area), lightgc[light_type], WhitePixel(display,screen_num));
	break;
      }
    }

    if (light->status != NEUTRAL)
    {  
      segments[0].x1 = light->x + light->radius - 4;
      segments[0].y1 = light->y;
      segments[0].x2 = light->x - light->radius + 4;
      segments[0].y2 = light->y;

      convertSegments(segments, 1);
      XDrawSegments(XtDisplay(drawing_area), plane, lightgc[light->type], segments, 1);

      if (light->status == POSITIVE)
      {
	segments[0].x1 = light->x;
	segments[0].y1 = light->y - light->radius + 4;
	segments[0].x2 = light->x;
	segments[0].y2 = light->y + light->radius - 4;

	convertSegments(segments, 1);
	XDrawSegments(XtDisplay(drawing_area), plane, lightgc[light->type], segments, 1);
      }
    }

    if (draw)
    {
      XCopyArea(XtDisplay(drawing_area), plane, XtWindow(drawing_area), lightgc[light->type], x, y,
		light->radius*2+4, light->radius*2+4, x, y);
    }

    p = p->next;
  }
}

//-----------------------------------------------------------------------------
// This function displays the age of the robot.
//-----------------------------------------------------------------------------
void ScreenUtility::displayAge()
{
  static double last_displayed_age = 0.0;
  char          msg[100];
  
  // Erase last display
  sprintf(msg, "Age    :%6.0f", last_displayed_age);
  drawText(gc, strdup(msg), 20, 20, ERASE);

  // Display new age
  sprintf(msg, "Age    :%6.0f", age);
  drawText(gc, strdup(msg), 20, 20, DRAW);

  last_displayed_age = age;
}

//-----------------------------------------------------------------------------
// This function displays the trials the experiment.
//-----------------------------------------------------------------------------
void ScreenUtility::displayTrials(void)
{
  static int      last_trial = 1;
  char            msg[100];

  // Erase last display
  if (trial != last_trial)
  { 
    sprintf(msg, "Trial  :%6d", trial);
    drawText(gc, strdup(msg), 20, 35, ERASE);
  }

  // Display new age
  sprintf(msg, "Trial  :%6d", trial);
  drawText(gc, strdup(msg), 20, 35, DRAW);

  last_trial = trial;
}

//-----------------------------------------------------------------------------
// This function displays the value for the robot's Vicarious Trial and Error
// (VTE).
//-----------------------------------------------------------------------------
void ScreenUtility::displayVTE(void)
{
  static int      last_VTE = 0;
  char            msg[100];

  t_node *pr = robotList;

  while (pr->next != NULL)
  {
    t_robot *robot = (t_robot *)pr->data;

    // Erase last display
    sprintf(msg, "VTE    :%6d", last_VTE);
    drawText(gc, strdup(msg), 20, 50, ERASE);

    // Display new age
    sprintf(msg, "VTE    :%6d", robot->vte);
    drawText(gc, strdup(msg), 20, 50, DRAW);
    last_VTE = robot->vte;

    pr = pr->next;
  }
}

//-----------------------------------------------------------------------------
// This function draws the text on the display.
//-----------------------------------------------------------------------------
void ScreenUtility::drawText(GC gc, char *text, int x, int y, int draw)
{
  int len = strlen(text);

  switch (draw) {
  case DRAW:
    XSetForeground(XtDisplay(drawing_area), gc, BlackPixel(XtDisplay(drawing_area), DefaultScreen(display)));
    break;
  case ERASE:
    XSetForeground(XtDisplay(drawing_area), gc, WhitePixel(XtDisplay(drawing_area), DefaultScreen(display)));
    break;
  }

  XDrawString(XtDisplay(drawing_area), XtWindow(drawing_area), gc, x, y, text, len);

  free(text);
}

//-----------------------------------------------------------------------------
// This function draws the box for the secondary robot window.
//-----------------------------------------------------------------------------
void ScreenUtility::drawWindow2box(Drawable win, GC gc)
{
  XSegment windowbox[4];

  windowbox[0].x1 = WINDOW2_ROBOT_X - WINDOW2_W/2;
  windowbox[0].y1 = WINDOW2_ROBOT_Y - WINDOW2_W/2;
  windowbox[0].x2 = WINDOW2_ROBOT_X + WINDOW2_W/2;
  windowbox[0].y2 = WINDOW2_ROBOT_Y - WINDOW2_W/2;

  windowbox[1].x1 = WINDOW2_ROBOT_X + WINDOW2_W/2;
  windowbox[1].y1 = WINDOW2_ROBOT_Y - WINDOW2_W/2;
  windowbox[1].x2 = WINDOW2_ROBOT_X + WINDOW2_W/2;
  windowbox[1].y2 = WINDOW2_ROBOT_Y + WINDOW2_W/2;

  windowbox[2].x1 = WINDOW2_ROBOT_X + WINDOW2_W/2;
  windowbox[2].y1 = WINDOW2_ROBOT_Y + WINDOW2_W/2;
  windowbox[2].x2 = WINDOW2_ROBOT_X - WINDOW2_W/2;
  windowbox[2].y2 = WINDOW2_ROBOT_Y + WINDOW2_W/2;

  windowbox[3].x1 = WINDOW2_ROBOT_X - WINDOW2_W/2;
  windowbox[3].y1 = WINDOW2_ROBOT_Y + WINDOW2_W/2;
  windowbox[3].x2 = WINDOW2_ROBOT_X - WINDOW2_W/2;
  windowbox[3].y2 = WINDOW2_ROBOT_Y - WINDOW2_W/2;

  convertSegments(windowbox, 4);
  XDrawSegments(XtDisplay(drawing_area), win, gc, windowbox, 4);
}

//-----------------------------------------------------------------------------
// This function creates the Pixmap from the bitmap.
//-----------------------------------------------------------------------------
Pixmap ScreenUtility::createColoredPixmap(int width, int height, unsigned char *bits, Pixel fg, Pixel bg)
{
  Pixmap pm = XCreatePixmapFromBitmapData(XtDisplay(top_level),
					  RootWindowOfScreen(XtScreen(top_level)),
					  (char *)bits,
					  width,
					  height,
					  fg,
					  bg,
					  DefaultDepthOfScreen(XtScreen(top_level)));
  return pm;
}

//-----------------------------------------------------------------------------
// This function creates the pixmap for the push buttons.
//-----------------------------------------------------------------------------
void ScreenUtility::buildBitMaps(Pixel fg, Pixel bg)
{
  pm_run_btn = createColoredPixmap(run_width, run_height, run_bits, fg, bg);
  pm_pause_btn = createColoredPixmap(pause_width, pause_height, pause_bits, fg, bg);
  pm_reset_btn = createColoredPixmap(reset_width, reset_height, reset_bits, fg, bg);
  pm_quit_btn = createColoredPixmap(quit_width, quit_height, quit_bits, fg, bg);
  pm_loadlights_btn = createColoredPixmap(loadlights_width, loadlights_height, loadlights_bits, fg, bg);
}

//-----------------------------------------------------------------------------
// This function moves the points in the XSegment for x and y.
//-----------------------------------------------------------------------------
void ScreenUtility::translateSegments(XSegment segments[], int nsegments, double x, double y)
{
  int i;
  for(i=0;i<nsegments;i++){
    segments[i].x1 += (int)x;
    segments[i].y1 += (int)y;
    segments[i].x2 += (int)x;
    segments[i].y2 += (int)y;
  }
}

//-----------------------------------------------------------------------------
// This function rotates the points in the XSegment around the center point
// (cx, cy).
//-----------------------------------------------------------------------------
void ScreenUtility::rotateSegments(double theta, XSegment segments[], int nsegments, double cx, double cy)
{

  int i;
  double stheta,ctheta;
  double x;

  theta = theta * M_PI / 180;
  stheta = sin(theta);
  ctheta = cos(theta);
  
  translateSegments(segments,nsegments,-cx,-cy);
   
  for(i=0;i<nsegments;i++){
    
    x = ctheta*(double)(segments[i].x1)-stheta*(double)(segments[i].y1);
    segments[i].y1 = (short int)(stheta*(double)(segments[i].x1)+ctheta*(double)(segments[i].y1));
    segments[i].x1 = (short int)x;
    x = ctheta*(double)(segments[i].x2)-stheta*(double)(segments[i].y2);
    segments[i].y2 = (short int)(stheta*(double)(segments[i].x2)+ctheta*(double)(segments[i].y2));
    segments[i].x2 = (short int)x;
  }
  
  translateSegments(segments,nsegments,cx,cy);
}

//-----------------------------------------------------------------------------
// This function flips the y direction of the XSegment.
//-----------------------------------------------------------------------------
void ScreenUtility::convertSegments(XSegment segments[], int nsegments)
{

  for(int i=0;i<nsegments;i++){
    
    segments[i].y1 = height - segments[i].y1;
    segments[i].y2 = height - segments[i].y2;
  }
}

/**********************************************************************
 * $Log: screen.cc,v $
 * Revision 1.42  2000/10/25 04:45:27  endo
 * Code cleaned up for the distribution.
 *
 * Revision 1.41  2000/08/19 20:46:29  endo
 * Slight changes due to the hardware interface.
 *
 * Revision 1.40  2000/08/07 21:01:19  endo
 * Reset feature added.
 *
 * Revision 1.39  2000/08/07 02:14:22  endo
 * Minor fix on Orientation Need window.
 *
 * Revision 1.38  2000/08/07 01:41:38  endo
 * Orientation Need Window created.
 *
 * Revision 1.37  2000/08/07 00:54:02  endo
 * ScreenUtility::simulateRobots changed as simbugMainLoop.
 *
 * Revision 1.36  2000/08/07 00:45:18  endo
 * trail improved.
 *
 * Revision 1.35  2000/08/06 07:26:08  endo
 * Two light types implemented successfully.
 *
 * Revision 1.34  2000/08/05 21:36:10  endo
 * TYPE_B color added.
 *
 * Revision 1.33  2000/08/05 19:40:35  endo
 * callbacks functions moved to callbacks.cc.
 *
 * Revision 1.32  2000/07/31 05:29:53  endo
 * Hypothesis feature improved. Orientation Need feature added.
 *
 * Revision 1.31  2000/07/29 22:17:08  endo
 * Reward stuff implemented.
 *
 * Revision 1.30  2000/04/30 23:56:31  endo
 * Finally, Internal Graph seems to be working correctly.
 *
 * Revision 1.29  2000/04/30 12:41:04  endo
 * Internal Graph implemented. Not perfect, though.
 *
 * Revision 1.28  2000/04/30 08:03:21  endo
 * photo_sensor implemented.
 *
 * Revision 1.27  2000/04/30 04:19:30  endo
 * color organized.
 *
 * Revision 1.26  2000/04/30 03:49:33  endo
 * shape of sowbug drawn correctly now.
 *
 * Revision 1.25  2000/04/30 01:55:14  endo
 * vectors are in color now.
 *
 * Revision 1.24  2000/04/30 01:35:00  endo
 * ArrowHeads success.
 *
 * Revision 1.23  2000/04/29 23:12:51  endo
 * drawTrail worked.
 *
 * Revision 1.22  2000/04/29 09:45:54  endo
 * Made pushRunPauseButton not to run simulation if there
 * is no light the field.
 *
 * Revision 1.21  2000/04/29 09:17:25  endo
 * Progression Vector working.
 *
 * Revision 1.20  2000/04/29 07:31:41  endo
 * Hypothesis Window added.
 *
 * Revision 1.19  2000/04/29 04:32:56  endo
 * orientation arrows show up correctly now.
 *
 * Revision 1.18  2000/04/28 09:19:45  endo
 * vector.value only show up in right side...
 *
 * Revision 1.17  2000/04/28 05:36:58  endo
 * line width experimented.
 *
 * Revision 1.16  2000/04/28 04:38:32  endo
 * secondary robot implemented.
 *
 * Revision 1.15  2000/04/27 21:26:58  endo
 * buttons added.
 *
 * Revision 1.14  2000/03/27 00:48:34  endo
 * Trying to implement drawVectors.
 *
 * Revision 1.13  2000/03/26 21:46:26  endo
 * ScreenUtility::xevent(void) improved.
 * By making screen_utility.simulateRobots to be
 * the XtAppAddWorkProc, the simulation doesn't get
 * stopped any more.
 *
 * Revision 1.12  2000/03/26 10:14:09  endo
 * drawSensors added.
 *
 * Revision 1.11  2000/03/26 09:55:47  endo
 * Placing lights from mouse clicks were also
 * successful.
 *
 * Revision 1.10  2000/03/26 08:34:58  endo
 * Finally, the simulation works as it supposed to be.
 *
 * Revision 1.9  2000/03/26 05:17:08  endo
 * Got rid of Start. Run/Pause instead.
 *
 * Revision 1.8  2000/03/13 06:37:21  endo
 * Start and Pause works. Getting right input (e.g., key
 * board char, mouse etc.) is still having a problem.
 *
 * Revision 1.7  2000/03/13 04:14:03  endo
 * The combination of XCheckMaskEvent and XtDispatchEvent
 * seems to be working.
 *
 * Revision 1.6  2000/03/12 09:10:28  endo
 * got rid of pthread, since it doesn't do any
 * good. XtAppNextEvent(app, &event) and
 * XtDispatchEvent (&event) applied. The
 * robot moves only when the mouse is moving.
 * timer etc. needed.
 *
 * Revision 1.5  2000/03/12 01:36:12  endo
 * kb_input moved to screen.cc
 *
 * Revision 1.4  2000/03/05 05:59:31  endo
 * pthread implemented. keyboard vs running robot is
 * the issue to be solved.
 *
 * Revision 1.3  2000/03/03 08:29:10  endo
 * Lot's of widgets (Motif) added.
 *
 * Revision 1.2  2000/03/03 02:11:01  endo
 * createMenuBar added. rcs tags added.
 *
 *
 **********************************************************************/
